// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"fmt"
	"os"

	"github.com/spf13/cobra"
	"github.com/spf13/viper"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/tools"
)

// accountCmd represents the account command.
var accountCmd = &cobra.Command{
	Use:   "account",
	Short: "Manage accounts that are allowed to connect to Xbus.",
	Long: `This command contains several sub-commands to manage accounts and the
permissions they have within Xbus.

In particular, it is possible from this interface to accept / reject pending
account requests.`,
}

// accountListCmd represents the account list sub-command.
var accountListCmd = &cobra.Command{
	Use:   "list",
	Short: "List accounts.",
	Long:  "Display accounts allowed to connect to Xbus, as a list.",
	Run:   listAccounts,
}

// accountAcceptCmd represents the account accept sub-command.
var accountAcceptCmd = &cobra.Command{
	Use:   "accept <fingerprint>",
	Short: "Accept an account.",
	Long:  "Accept a pending account request so it can be allowed to connect to Xbus.",
	Run:   acceptAccount,
}

// accountRejectCmd represents the account reject sub-command.
var accountRejectCmd = &cobra.Command{
	Use:   "reject <fingerprint>",
	Short: "Reject an account.",
	Long:  "Reject a pending account request; it will not be allowed to connect to Xbus.",
	Run:   rejectAccount,
}

// Arguments of the "accept" command.
const (
	acceptCmdArgFingerprint = iota
	acceptCmdArgCount
)

// Arguments of the "reject" command.
const (
	rejectCmdArgFingerprint = iota
	rejectCmdArgCount
)

// listAccounts parses arguments to list accounts.
func listAccounts(cmd *cobra.Command, args []string) {
	csrHash := viper.GetBool("csr-hash")
	certDetails := viper.GetBool("cert-details")
	os.Exit(doListAccounts(csrHash, certDetails))
}

// doListAccounts lists accounts.
func doListAccounts(csrHash bool, certDetails bool) int {
	conn, err := GetClient()
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	accounts, err := conn.AccountList()
	if err != nil {
		Logger.Errorf("Cannot retrieve accounts: %s", err)
		return 1
	}

	accountCount := len(accounts)
	if accountCount == 0 {
		fmt.Println("No account found.")
	} else {
		fmt.Printf("%d accounts found:\n", accountCount)
		for _, account := range accounts {
			fmt.Println(fmtAccount(account, csrHash, certDetails))
		}
	}
	return 0
}

// acceptAccount parses arguments to accept a pending account request.
func acceptAccount(cmd *cobra.Command, args []string) {
	if len(args) < acceptCmdArgCount {
		Logger.Errorf("Not enough arguments; %d arguments are required.", acceptCmdArgCount)
		cmd.Usage()
		os.Exit(1)
	}

	os.Exit(doAcceptAccount(args[acceptCmdArgFingerprint]))
}

// doAcceptAccount accepts a pending account request.
func doAcceptAccount(fingerprint string) int {
	conn, err := GetClient()
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	account, found := findAccountFromCSRFingerprint(conn, fingerprint)
	if !found {
		return 1
	}

	// Got a match! Send an "accept" request.
	repAccount, err := conn.AccountAccept(account.ID)
	if err != nil {
		Logger.Errorf("Could not accept the account %q: %s", account.Name, err)
		return 1
	}
	if repAccount.Status == api.AccountStatusPending {
		Logger.Errorf("Could not accept the account %q: Still pending.", account.Name)
		return 1
	}
	Logger.Noticef("Successfully accepted the account %q.", repAccount.Name)
	return 0
}

// rejectAccount parses arguments to reject a pending account request.
func rejectAccount(cmd *cobra.Command, args []string) {
	if len(args) < rejectCmdArgCount {
		Logger.Errorf("Not enough arguments; %d arguments are required.", rejectCmdArgCount)
		cmd.Usage()
		os.Exit(1)
	}

	os.Exit(doRejectAccount(args[rejectCmdArgFingerprint]))
}

// doRejectAccount rejects a pending account request.
func doRejectAccount(fingerprint string) int {
	conn, err := GetClient()
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	account, found := findAccountFromCSRFingerprint(conn, fingerprint)
	if !found {
		return 1
	}

	// Got a match! Send a "reject" request.
	repAccount, err := conn.AccountReject(account.ID)
	if err != nil {
		Logger.Errorf("Could not reject the account %q: %s", account.Name, err)
		return 1
	}
	if repAccount.Status == api.AccountStatusPending {
		Logger.Errorf("Could not reject the account %q: Still pending.", account.Name)
		return 1
	}
	Logger.Noticef("Successfully rejected the account %q.", repAccount.Name)
	return 0
}

// findAccountFromCSRFingerprint finds a pending account with the specified fingerprint.
func findAccountFromCSRFingerprint(
	conn *xbus.Client, fingerprint string,
) (
	account api.Account, found bool,
) {
	// TODO To be done elsewhere? Should fingerprints be cached?

	accounts, err := conn.AccountList()
	if err != nil {
		Logger.Errorf("Cannot retrieve accounts: %s", err)
		return api.Account{}, false
	}

	for _, account := range accounts {
		if tools.GetCSRHash(account.CSR, true) == fingerprint {
			// Got a match! Ensure the account is pending.
			if account.Status != api.AccountStatusPending {
				Logger.Errorf("The specified account (%s) is not pending a validation.",
					account.Name)
				return api.Account{}, false
			}
			return account, true
		}
	}

	Logger.Errorf("No account found with the specified fingerprint (%s).", fingerprint)
	return api.Account{}, false
}

func fmtAccount(account api.Account, csrHash bool, certDetails bool) string {
	formattedAccount := fmt.Sprintf(`
ID: %s
Name: %s
Type: %s
Status: %s`,
		account.ID,
		account.Name,
		account.Type,
		account.Status,
	)
	if csrHash {
		formattedAccount += fmt.Sprintf(`
CSR fingerprint: %s`,
			tools.GetCSRHash(account.CSR, true),
		)
	}
	if certDetails {
		formattedAccount += fmt.Sprintf(`
CSR: %s
Cert: %s`,
			account.CSR,
			account.Cert,
		)
	}
	return formattedAccount
}

func init() {
	RootCmd.AddCommand(accountCmd)

	accountCmd.AddCommand(accountListCmd)
	accountListFlags := accountListCmd.PersistentFlags()
	accountListFlags.Bool("cert-details", false, "Display cert details")
	viper.BindPFlag("cert-details", accountListFlags.Lookup("cert-details"))
	accountListFlags.Bool("csr-hash", true, "Display CSR fingerprints")
	viper.BindPFlag("csr-hash", accountListFlags.Lookup("csr-hash"))

	accountCmd.AddCommand(accountAcceptCmd)

	accountCmd.AddCommand(accountRejectCmd)
}
