// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"fmt"
	"io/ioutil"
	"os"

	"github.com/spf13/cobra"
	"gopkg.in/yaml.v2"

	"bitbucket.org/orus-io/xbus/storage"
)

func graphQuery(cmd *cobra.Command, args []string) {
	c, err := GetClient()
	if err != nil {
		return
	}
	name := cmd.Flag("name").Value.String()
	save := cmd.Flag("save").Value.String() == "true"
	all := cmd.Flag("all").Value.String() == "true"

	graphs, err := c.GraphQuery(name, !all)
	if err != nil {
		Logger.Errorf("%s", err)
		return
	}
	for _, graph := range graphs {
		var sGraph storage.Graph
		if err := yaml.Unmarshal([]byte(graph), &sGraph); err != nil {
			Logger.Errorf("Cannot read graph: %s", err)
			return
		}
		if save {
			if err := ioutil.WriteFile(
				sGraph.Name+"-"+sGraph.Version+".yml",
				[]byte(graph), 0660); err != nil {
				Logger.Errorf("Cannot save the graph: %s", err)
				return
			}
		}
	}
}

func graphPush(cmd *cobra.Command, args []string) {
	if len(args) < pushCmdArgCount {
		Logger.Errorf("Not enough arguments; %d arguments are required.", pushCmdArgCount)
		cmd.Usage()
		os.Exit(1)
	}

	c, err := GetClient()
	if err != nil {
		return
	}

	filename := args[pushCmdArgFilepath]

	content, err := ioutil.ReadFile(filename)
	if err != nil {
		Logger.Errorf("Cannot read graph file: %s", err)
		return
	}

	graph, validationMessages, err := c.GraphPush(string(content))
	if len(validationMessages) != 0 {
		fmt.Println("Validation messages:")
		for _, msg := range validationMessages {
			fmt.Println("-", msg)
		}
	}
	if err != nil {
		Logger.Errorf("%s", err)
		return
	}

	if err := ioutil.WriteFile(filename, []byte(graph), 0660); err != nil {
		Logger.Errorf("Cannot save the resulting graph: %s", err)
		return
	}
}

var graphCmd = &cobra.Command{
	Use:   "graph",
	Short: "Graph control subcommands",
	Long:  "Graph subcommands allow to list and modify graphs",
}

var graphQueryCmd = &cobra.Command{
	Use:   "query",
	Short: "Search for graphs",
	Run:   graphQuery,
}

var graphPushCmd = &cobra.Command{
	Use:   "push <filepath>",
	Short: "Push a graph",
	Run:   graphPush,
}

// Arguments of the "push" command.
const (
	pushCmdArgFilepath = iota
	pushCmdArgCount
)

func init() {
	RootCmd.AddCommand(graphCmd)
	graphCmd.AddCommand(graphQueryCmd)
	graphCmd.AddCommand(graphPushCmd)

	flags := graphQueryCmd.PersistentFlags()
	flags.String("name", "", "graph name")
	flags.Bool("save", false, "save graphs to disk")
	flags.Bool("all", false, "query all graphs, not only the active ones")
}
