// Copyright 2016 Orus.io. All rights reserved.

package fullstack

import (
	"crypto/ecdsa"
	"crypto/elliptic"
	"crypto/rand"
	"crypto/x509"
	"crypto/x509/pkix"
	"encoding/pem"
	"fmt"
	"math/big"
	"testing"
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/nats-io/go-nats"
	"github.com/orus-io/yago"
	"github.com/stretchr/testify/assert"
	// "golang.org/x/crypto/ed25519"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/server"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
	"bitbucket.org/orus-io/xbus/testutils"
	"bitbucket.org/orus-io/xbus/testutils/sqldb"
)

// Fullstack is the stub we use in our tests
type Fullstack struct {
	T *testing.T

	ServerCert []byte

	Server *server.Server

	Client *xbus.Client

	Logger  *testutils.TestLogger
	Model   *model.Model
	DB      *yago.DB
	CloseDB sqldb.CloseFunction
}

func getCertTemplate(ca bool) x509.Certificate {
	serialNumberLimit := new(big.Int).Lsh(big.NewInt(1), 128)
	serialNumber, err := rand.Int(rand.Reader, serialNumberLimit)
	if err != nil {
		panic(err)
	}

	template := x509.Certificate{
		SerialNumber: serialNumber,
		Subject: pkix.Name{
			Organization: []string{"Test"},
			CommonName:   "localhost", // Tests are all done on localhost
		},
		NotBefore: time.Now(),
		NotAfter:  time.Now().Add(1 * time.Minute),

		IsCA:        ca,
		KeyUsage:    x509.KeyUsageKeyEncipherment | x509.KeyUsageDigitalSignature | x509.KeyUsageCertSign,
		ExtKeyUsage: []x509.ExtKeyUsage{x509.ExtKeyUsageServerAuth, x509.ExtKeyUsageAny},

		BasicConstraintsValid: true,
	}
	return template
}

// InitFullstack is used to produce and kickstart a full stack environment
// ready for testing
func InitFullstack(t *testing.T, autostart bool) *Fullstack {
	privateKey, err := ecdsa.GenerateKey(elliptic.P256(), rand.Reader)
	if err != nil {
		t.Error(err)
		return nil
	}

	// create a cert template for our Root CA cert
	template := getCertTemplate(true)
	// and generate the cert
	bCert, err := x509.CreateCertificate(
		rand.Reader, &template, &template, &privateKey.PublicKey, privateKey)
	if err != nil {
		t.Error(err)
		return nil
	}

	derKey, err := x509.MarshalECPrivateKey(privateKey)
	if err != nil {
		t.Error(err)
		return nil
	}
	pemKey := string(pem.EncodeToMemory(&pem.Block{
		Type:  "EC PRIVATE KEY",
		Bytes: derKey,
	}))

	pemCert := string(pem.EncodeToMemory(&pem.Block{
		Type:  "CERTIFICATE",
		Bytes: bCert,
	}))

	m, db, closeDB := sqldb.GetTestDB(t)
	logger := testutils.GetTestLogger(t)

	s := server.NewServer(server.Options{
		RootCA:                   pemCert,
		ServerCert:               pemCert,
		ServerKey:                pemKey,
		ClientCA:                 pemCert,
		ClientCAKey:              pemKey,
		NATSHost:                 "localhost",
		NATSPort:                 -1, // Let the system find an available port
		ServeNATS:                true,
		ServeBroker:              true,
		ServeEmission:            true,
		ServeControl:             true,
		ServeProcessState:        true,
		ServeRegistration:        true,
		ServeSentinel:            true,
		EmissionReceptionTimeout: 100 * time.Millisecond,
	})

	s.SetSQLDB(m, db)
	s.SetLogger(logger)

	ft := Fullstack{
		T: t,

		Logger: logger,

		ServerCert: bCert,
		Server:     s,
		Model:      m,
		DB:         db,
		CloseDB:    closeDB,
	}

	if !autostart || ft.ServerStart() {
		return &ft
	}
	ft.Shutdown()
	return nil
}

// ServerStart is responsible to start the Xbus server
func (ft *Fullstack) ServerStart() bool {
	if err := ft.Server.Startup(); err != nil {
		ft.T.Errorf("Failed to start server: %s", err)
		return false
	}

	cl := xbus.NewClient(
		xbus.COptions{ConnOptions: ft.Server.NATSConnectOptions()},
		ft.Logger,
	)

	err := cl.Connect()
	assert.Nil(ft.T, err)
	if err != nil {
		ft.Shutdown()
		return false
	}
	ft.Client = cl

	return true
}

// ServerStop is used during shutdown to properly exit all processes
func (ft *Fullstack) ServerStop() {
	if ft == nil {
		return
	}
	if ft.Client != nil {
		ft.Client.Close()
		ft.Client = nil
	}
	ft.Server.Shutdown()
}

// Shutdown is what you call to teardown the whole fullstack env
func (ft *Fullstack) Shutdown() {
	if ft == nil {
		return
	}
	ft.ServerStop()
	if ft.CloseDB != nil {
		ft.CloseDB()
		ft.CloseDB = nil
	}
	ft.Logger.VerifyExpectations()
}

// CreateRole ...
func (ft *Fullstack) CreateRole(actorKind api.ActorKind, name string) uuid.UUID {
	// no-op
	return uuid.UUID{}
}

// CreateAccount ...
func (ft *Fullstack) CreateAccount(name string, aType api.AccountType) *xbus.Client {
	c := xbus.NewClient(xbus.COptions{
		Name: name,
		Type: aType,
		ConnOptions: nats.Options{
			Url: ft.Server.NATSUrl(),
		},
		RootCAs: [][]byte{ft.ServerCert},
	}, ft.Logger)

	if err := c.Connect(); err != nil {
		panic(err)
	}

	if err := c.CreateKey(false); err != nil {
		ft.T.Fatal(err)
	}
	// generate a new CSR from our new private key
	c.CreateCSR("localhost", "DSI", "client.xbus.io", false)
	status, err := c.Register()
	if err != nil {
		panic(err)
	}
	assert.Equal(ft.T, api.RegistrationPending, status)

	account, err := ft.Client.AccountAccept(c.ID())
	if err != nil {
		ft.T.Fatal(err)
	}
	assert.Equal(ft.T, api.AccountStatusEnabled, account.Status)

	status, err = c.Register()
	assert.Equal(ft.T, api.RegistrationAccepted, status)

	c.Close()
	if err := c.Connect(); err != nil {
		ft.T.Fatal(err)
	}

	return c
}

func (ft *Fullstack) createActor(actorKind api.ActorKind, name string, roles []string) *xbus.Client {
	if len(roles) != 1 {
		panic("One and only one role supported for now")
	}

	// Create an account.
	ac := ft.CreateAccount(name, api.AccountTypeActor)

	ac.AddActor(xbus.ActorOptions{Actor: api.Actor{
		Name:  name,
		Kind:  actorKind,
		Roles: roles,
	}})

	if _, err := ac.Register(); err != nil {
		ft.T.Error(err)
		return nil
	}

	actorOptions, ok := ac.GetActorOptionsByName(name)
	assert.True(ft.T, ok)
	assert.Equal(ft.T, api.ActorPending, actorOptions.Status)

	_, err := ft.Client.ActorAccept(actorOptions.ID)
	if err != nil {
		ft.T.Error(err)
		return nil
	}

	if _, err := ac.Register(); err != nil {
		ft.T.Error(err)
		return nil
	}

	actorOptions, ok = ac.GetActorOptionsByName(name)
	assert.True(ft.T, ok)
	assert.Equal(ft.T, api.ActorEnabled, actorOptions.Status)

	ac.Close()
	// Reconnect the client so the new permissions are loaded
	assert.Nil(ft.T, ac.Connect())

	return ac
}

// CreateEmitter ...
func (ft *Fullstack) CreateEmitter(name string, roles []string) (*xbus.Client, *xbus.Actor) {
	c := ft.createActor(api.EmitterKind, name, roles)

	actor := c.GetActorByName(name)
	assert.NotNil(ft.T, actor)
	assert.Equal(ft.T, api.ActorEnabled, actor.Options.Status)

	return c, actor
}

// CreateConsumer ...
func (ft *Fullstack) CreateConsumer(name string, roles []string) (*xbus.Client, *xbus.Actor) {
	c := ft.createActor(api.ConsumerKind, name, roles)

	actor := c.GetActorByName(name)
	assert.NotNil(ft.T, actor)
	assert.Equal(ft.T, api.ActorEnabled, actor.Options.Status)

	return c, actor
}

// CreateWorker ...
func (ft *Fullstack) CreateWorker(name string, roles []string) (*xbus.Client, *xbus.Actor) {
	c := ft.createActor(api.WorkerKind, name, roles)

	actor := c.GetActorByName(name)
	assert.NotNil(ft.T, actor)
	assert.Equal(ft.T, api.ActorEnabled, actor.Options.Status)

	return c, actor
}

// BuildSimpleGraph builds a graph
func BuildSimpleGraph(eventtype, consumerrole string) string {
	return fmt.Sprintf(`name: %s -> %s
version: 0.1
status: active
nodes:
- id: source
  type: emitter
  sourcematch:
    eventtypes: [%s]
  outputs: [default]
- id: target
  type: consumer
  actorids: []
  roles: [%s]
  rolebroadcast: true
  inputs: [default]
edges:
- source -> target
`,
		eventtype, consumerrole,
		eventtype,
		consumerrole)
}
