package graph

import (
	"errors"

	"github.com/m4rw3r/uuid"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
)

// NewManager returns a Manager
func NewManager(
	actorStorage storage.ActorStorage,
	graphStorage storage.GraphStorage,
) Manager {
	return Manager{
		ActorStorage: actorStorage,
		GraphStorage: graphStorage,
	}
}

// Manager knows all the graphs
type Manager struct {
	ActorStorage storage.ActorStorage
	GraphStorage storage.GraphStorage
	Graphs       []Graph
}

func (m *Manager) loadActiveGraphs() {
	graphs := m.GraphStorage.LoadActiveGraph()

	for _, graph := range graphs {
		m.Graphs = append(m.Graphs, NewGraph(graph))
	}
}

// LookupGraph returns the matching graph for a specific event type
func (m *Manager) LookupGraph(emitter api.Actor, eventtypes []string) (Graph, storage.NodeOutputRef, bool) {
	if len(m.Graphs) == 0 {
		m.loadActiveGraphs()
	}
	for _, graph := range m.Graphs {
		output, match := graph.Match(emitter, eventtypes)
		if match {
			return graph, output, true
		}
	}
	return Graph{}, storage.NodeOutputRef{}, false
}

// GetGraph load a graph from its id
func (m *Manager) GetGraph(id uuid.UUID) (Graph, bool) {
	graph, exists := m.GraphStorage.Load(id)
	return NewGraph(graph), exists
}

// Save validates and saves a graph
func (m *Manager) Save(graph Graph) (Graph, error) {
	if graph.Status != storage.GraphDraft {
		err := graph.Validate()
		if len(err) != 0 {
			return Graph{}, errors.New(
				"Graph is not valid, cannot publish or activate it")
		}
	}
	if graph.Status == storage.GraphActive {
		err := graph.ValidateActors(m.ActorStorage)
		if err != nil {
			return Graph{}, errors.New(
				"Graph cannot be activated: some roles/actors are unkwown in this bus")
		}
	}
	id, err := m.GraphStorage.Save(graph.Graph)
	if err != nil {
		return Graph{}, err
	}
	g, ok := m.GetGraph(id)
	if !ok {
		panic("Saved graph is missing")
	}
	return g, nil
}

// Query lookup for graphs match some criterions
func (m *Manager) Query(active bool) (result []Graph) {
	var query storage.Graph
	if active {
		query.Status = storage.GraphActive
	}
	for _, graph := range m.GraphStorage.Query(query) {
		result = append(result, NewGraph(graph))
	}
	return
}
