package server

import (
	"crypto/ecdsa"
	"crypto/tls"
	"crypto/x509"
	"errors"
	"fmt"
	"strconv"
	"strings"
	"time"

	gnatsd "github.com/nats-io/gnatsd/server"
	"github.com/nats-io/go-nats"
	"github.com/orus-io/yago"
	"github.com/slicebit/qb"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/tools"
	"bitbucket.org/orus-io/xbus/auth"
	"bitbucket.org/orus-io/xbus/graph"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/storage/sql"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
)

// Options of a server
type Options struct {
	DatabaseDSN       string
	DatabaseLogFlags  qb.LogFlags
	RootCA            string // a PEM or a path to a PEM file
	ServerCert        string // a PEM or a path to a PEM file
	ServerKey         string // a PEM or a path to a PEM file
	ClientCA          string // a PEM or a path to a PEM file
	ClientCAKey       string // a PEM or a path to a PEM file
	NATSHost          string
	NATSPort          int
	NATSDebug         bool
	NATSTrace         bool
	ServeNATS         bool
	ServeBroker       bool
	ServeEmission     bool
	ServeControl      bool
	ServeProcessState bool
	ServeRegistration bool
	ServeSentinel     bool

	EmissionReceptionTimeout time.Duration
}

// Server the XBus server
type Server struct {
	options Options

	model *model.Model
	db    *yago.DB

	accountStorage  storage.AccountStorage
	actorStorage    storage.ActorStorage
	graphStorage    storage.GraphStorage
	processStorage  storage.ProcessStorage
	envelopeStorage storage.EnvelopeStorage
	logStorage      storage.LogStorage

	processManager *process.Manager
	graphManager   graph.Manager

	rootCA      *x509.Certificate
	serverCert  *x509.Certificate
	serverKey   *ecdsa.PrivateKey
	clientCA    *x509.Certificate
	clientCAKey *ecdsa.PrivateKey

	natsAuth   auth.XBusAuth
	natsServer *gnatsd.Server
	natsConn   *nats.EncodedConn

	ServiceManager *services.Manager

	log        xbus.Logger
	natsLogger xbus.Logger
}

// NewServer creates a Server
func NewServer(options Options) *Server {
	return &Server{
		options:    options,
		log:        xbus.NewNoopLogger(),
		natsLogger: xbus.NewNoopLogger(),
	}
}

// SetLogger changes the logger used by server
func (s *Server) SetLogger(logger xbus.Logger) {
	if logger == nil {
		s.log = xbus.NewNoopLogger()
		s.natsLogger = s.log
	} else {
		s.log = xbus.NewPrefixedLogger("[server] ", logger)
		s.natsLogger = xbus.NewPrefixedLogger("[gnatsd] ", logger)
	}
}

// SetSQLDB sets the model & db
func (s *Server) SetSQLDB(m *model.Model, db *yago.DB) {
	s.model = m
	s.db = db
}

func (s *Server) setupSQLDB() error {
	if s.model != nil {
		return nil
	}
	m, db, err := model.NewDB(
		"postgres", s.options.DatabaseDSN,
		xbus.NewPrefixedLogger("[sql] ", s.log),
		s.options.DatabaseLogFlags)
	if err != nil {
		s.log.Errorf("Cannot connect to the database.Got err %s", err)
		return err
	}
	s.SetSQLDB(m, db)
	return nil
}

// AccountStorage returns the underlying AccountStorage
func (s *Server) AccountStorage() storage.AccountStorage {
	return s.accountStorage
}

// ActorStorage returns the underlying ActorStorage
func (s *Server) ActorStorage() storage.ActorStorage {
	return s.actorStorage
}

// ProcessStorage returns the underlying ProcessStorage
func (s *Server) ProcessStorage() storage.ProcessStorage {
	return s.processStorage
}

func (s *Server) setupStorage() {
	s.accountStorage = sql.NewAccountStorage(s.model, s.db, s.log)
	s.actorStorage = sql.NewActorStorage(s.model, s.db, s.log)
	s.envelopeStorage = sql.NewEnvelopeStorage(s.model, s.db, s.log)
	s.graphStorage = sql.NewGraphStorage(s.model, s.db, s.log)
	s.processStorage = sql.NewProcessStorage(s.model, s.db, s.log)
	s.logStorage = sql.NewLogStorage(s.model, s.db, s.log)
}

func (s *Server) setupManagers() {
	s.graphManager = graph.NewManager(
		s.actorStorage, s.graphStorage)
	s.processManager = process.NewManager(
		s.graphManager, s.actorStorage, s.processStorage, s.log,
	)
}

func (s *Server) setupCrypto() error {
	rootCA, err := tools.ReadPEMCertificate(s.options.RootCA)
	if err != nil {
		return err
	}
	serverCert, err := tools.ReadPEMCertificate(s.options.ServerCert)
	if err != nil {
		return err
	}
	serverKey, err := tools.ReadPEMPrivateKey(s.options.ServerKey)
	if err != nil {
		return err
	}
	clientCA, err := tools.ReadPEMCertificate(s.options.ClientCA)
	if err != nil {
		return err
	}
	clientCAKey, err := tools.ReadPEMPrivateKey(s.options.ClientCAKey)
	if err != nil {
		return err
	}
	s.rootCA = rootCA
	s.serverCert = serverCert
	s.serverKey = serverKey
	s.clientCA = clientCA
	s.clientCAKey = clientCAKey

	return nil
}

// Startup starts the server
func (s *Server) Startup() (err error) {
	defer func() {
		if err != nil {
			s.Shutdown()
		}
	}()
	if err = s.setupSQLDB(); err != nil {
		return err
	}
	s.setupStorage()
	s.setupManagers()
	if err = s.setupCrypto(); err != nil {
		return err
	}
	if s.options.ServeNATS {
		if err = s.setupNATS(); err != nil {
			return err
		}
		if err = s.StartNATS(); err != nil {
			return err
		}
	}
	s.setupServices()
	s.log.Noticef("Starting services")
	if err = s.ServiceManager.Startup(); err != nil {
		return err
	}
	return nil
}

// Shutdown stops the services and nats server (if relevant)
func (s *Server) Shutdown() {
	if s.ServiceManager != nil {
		if err := s.ServiceManager.Shutdown(); err != nil {
			s.log.Errorf("Error shutting down services: %s", err)
		}
	}
	if s.options.ServeNATS {
		s.StopNATS()
	}
}

func (s *Server) setupNATS() error {
	certPool := x509.NewCertPool()
	certPool.AddCert(s.clientCA)

	s.natsAuth = auth.NewXBusAuth(s.accountStorage, s.actorStorage, s.log)

	s.natsServer = gnatsd.New(&gnatsd.Options{
		Host:   s.options.NATSHost,
		Port:   s.options.NATSPort,
		NoSigs: true,
		TLSConfig: &tls.Config{
			Certificates: []tls.Certificate{
				tls.Certificate{
					Certificate: [][]byte{s.serverCert.Raw},
					PrivateKey:  s.serverKey,
					Leaf:        s.serverCert,
				},
			},
			ClientAuth: tls.VerifyClientCertIfGiven,
			ClientCAs:  certPool,
		},
	})
	s.natsServer.SetClientAuthMethod(&s.natsAuth)
	s.natsServer.SetLogger(s.natsLogger, s.options.NATSDebug, s.options.NATSTrace)

	return nil
}

// getNATSPort returns the nats server port, when an
// error occurs the returned value should be ignored.
func (s *Server) getNATSPort() (int, error) {
	endpoint := s.natsServer.Addr()
	if endpoint == nil {
		return 0, errors.New("Undefined endpoint")
	}
	subStr := strings.SplitAfter(endpoint.String(), ":")
	strPort := subStr[len(subStr)-1]
	port, err := strconv.Atoi(strPort)
	if err != nil {
		return 0, err
	}
	return port, nil
}

// NATSUrl returns the nats server url
func (s *Server) NATSUrl() string {
	port, err := s.getNATSPort()
	if err != nil {
		panic(err)
	}
	return fmt.Sprintf("nats://%s:%d", s.options.NATSHost, port)
}

// NATSConnectOptions returns nats.Options to connect to the
// bus with full access
func (s *Server) NATSConnectOptions() nats.Options {
	certPool := x509.NewCertPool()
	certPool.AddCert(s.rootCA)

	return nats.Options{
		Url:   s.NATSUrl(),
		Token: s.natsAuth.AuthToken,
		TLSConfig: &tls.Config{
			RootCAs: certPool,
		},
		Secure: true,
	}
}

func (s *Server) waitForNATS() error {

	if !s.natsServer.ReadyForConnections(time.Second) {
		return errors.New("Server still has no listen endpoint")
	}
	url := s.NATSUrl()
	if url == "" {
		return errors.New("Server still has no listen endpoint")
	}

	/*
		opts := s.NATSConnectOptions()

		var (
			ready   bool
			lastErr error
		)
		for time.Now().Before(timeout) {
			// let the goroutine start to listen
			runtime.Gosched()

			natsConn, err := opts.Connect()
			lastErr = err
			if err == nil {
				natsConn.Close()
				ready = true
				break
			}
		}
		if !ready {
			return fmt.Errorf("Cannot connect to the nats server: %s", lastErr)
		}
	*/
	s.log.Noticef("Nats server responding on %s", s.NATSUrl())

	// Store the actual port back into the options so that a restart
	// will reuse the same port, even if the original value is -1 which means
	// a random port.
	port, err := s.getNATSPort()
	if err != nil {
		panic(err)
	}
	s.options.NATSPort = port
	return nil
}

// StartNATS starts the nats server
func (s *Server) StartNATS() error {
	s.log.Noticef("Starting NATS server")

	go s.natsServer.Start()

	if err := s.waitForNATS(); err != nil {
		s.StopNATS()
		return err
	}

	connOptions := nats.Options{
		Url:    s.NATSUrl(),
		Token:  s.natsAuth.AuthToken,
		Secure: true,
		AsyncErrorCB: func(c *nats.Conn, sub *nats.Subscription, err error) {
			if sub != nil {
				s.log.Errorf("Error on sub %s: %s", sub.Subject, err)
			} else {
				s.log.Errorf("Error %s", err)
			}
		},
	}

	conn, err := connOptions.Connect()

	if err != nil {
		s.log.Errorf("Could not connect to the NATS server: %s", err)
		s.StopNATS()
		return err
	}

	eConn, err := nats.NewEncodedConn(conn, "json")

	if err != nil {
		conn.Close()
		s.StopNATS()
		return fmt.Errorf("Could not initialize the encoded connection: %s", err)
	}

	s.natsConn = eConn

	return nil
}

// StopNATS stops the nats server
func (s *Server) StopNATS() {
	if s.natsConn != nil {
		s.natsConn.Close()
		s.natsConn = nil
	}
	if s.natsServer != nil {
		s.natsServer.Shutdown()
		s.natsServer = nil
	}
}
