// Copyright 2017 Orus.io. All rights reserved.

package broker

import (
	"fmt"
	"strings"

	"github.com/m4rw3r/uuid"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/storage"
)

func parseSubject(subject string) (actorID uuid.UUID, output string, err error) {
	splitted := strings.Split(subject, ".")
	if len(splitted) != 5 {
		return uuid.UUID{}, "", fmt.Errorf("Invalid subject. Expected 5 parts, got %d", len(splitted))
	}
	if splitted[0] != "xbus" || splitted[1] != "msgbox" || splitted[3] != "output" {
		return uuid.UUID{}, "", fmt.Errorf("Invalid subject. Expected xbus.msgbox.*.output.*, got %s", subject)
	}
	actorID, err = uuid.FromString(splitted[2])
	if err != nil {
		return
	}
	output = splitted[4]
	return
}

func (s *Service) replyAckError(reply string, envelopeID uuid.UUID, err error) {
	s.Log.Errorf("error: %s", err)
	err = s.Conn.Publish(reply, api.NewEnvelopeAckError(envelopeID, err))
	if err != nil {
		s.Log.Errorf("Error replying: %s", err)
	}
	return
}

func (s *Service) handleIncomingEnvelope(subject, reply string, envelope api.Envelope) {
	s.Log.Noticef("Handling incoming envelope")
	var ack = api.EnvelopeAck{ID: envelope.ID}
	s.TaskAdd()
	defer s.SendReply(reply, &ack)

	actorID, output, err := parseSubject(subject)
	if err != nil {
		ack.Response = api.ClientError("Invalid subject '%s': %s", subject, err)
		return
	}

	actor, exists := s.Opts.ActorStorage.Get(actorID)
	if !exists {
		ack.Response = api.ClientError("Invalid actor: %s", actorID)
		return
	}

	if envelope.Context.ProcessID.IsZero() && actor.Kind != api.EmitterKind {
		s.Log.Errorf("Context-less envelope sent by non-emitter %s", actor.Name)
		ack.Response = api.ClientError("Context-less envelope sent by non-emitter %s", actor.Name)
		return
	}

	var (
		process   *process.Process
		sourceRef storage.NodeOutputRef
	)
	if !envelope.Context.ProcessID.IsZero() {
		process, err = s.processManager.Get(envelope.Context.ProcessID)
		if err != nil {
			err := fmt.Errorf("Invalid process on incoming envelope. Rejecting it")
			s.Log.Errorf("%s", err)
			err = s.Conn.Publish(reply, api.NewEnvelopeAckError(envelope.ID, err))
			return
		}
		sourceRef.NodeID = envelope.Context.NodeID
		sourceRef.Output = output
	} else {
		// TODO enforce sender is an autorized emitter
	}

	var (
		receptionStatus api.ReceptionStatus
		statusReason    string
		typesKnown      bool
		eventTypes      []string
	)

	if envelope.Last == true && len(envelope.Events) == 0 && len(envelope.EventIDs) == 0 {
		if process == nil {
			s.replyAckError(
				reply, envelope.ID,
				fmt.Errorf("Invalid envelop (missing or invalid context)"))
			return
		}

		// empty envelope, which can mean 2 things depending on the ID

		// Without an ID: the envelope is just a signal that the output
		// will not be used in this context. Inform the process.
		if envelope.ID.IsZero() {
			err = s.Conn.Publish(
				reply,
				api.EnvelopeAck{ID: envelope.ID, ReceptionStatus: api.ReceptionAccepted},
			)
			process.UpdateReceiveStatus(
				envelope.ID, actor.ID, sourceRef, storage.ReceptionClosed)
			if err != nil {
				s.Log.Errorf("Error replying: %s", err)
			}
			return
		}
		// With an ID: the envelope is one sent on an input, hence must be
		// known to the broker which can reread it from storage.
		receptionStatus = api.ReceptionAccepted
		typesKnown = true
	} else {
		status, err := s.Opts.EnvelopeStorage.StoreEnvelope(envelope)
		if err != nil {
			s.replyAckError(reply, envelope.ID, err)
			return
		}
		s.Log.Noticef("Stored envelope: %s", status.Status)
		switch status.Status {
		case "receiving":
			receptionStatus = api.ReceptionReceiving
		case "complete":
			receptionStatus = api.ReceptionAccepted
		case "error":
			receptionStatus = api.ReceptionError
		default:
			receptionStatus = api.ReceptionError
		}
		typesKnown = status.TypesKnown
		eventTypes = status.EventTypes()

		if receptionStatus == api.ReceptionError {
			if process != nil {
				process.UpdateReceiveStatus(
					envelope.ID, actor.ID, sourceRef, storage.ReceptionError)
			}
			err = s.Conn.Publish(reply, api.NewEnvelopeAckError(envelope.ID, status))
			if err != nil {
				s.Log.Errorf("Error replying: %s", err)
			}
			return
		}
	}

	if envelope.Context.ProcessID.IsZero() && typesKnown {
		process, err = s.processManager.LookupOrCreate(actor, envelope.ID, eventTypes)
		if err != nil {
			receptionStatus = api.ReceptionError
			statusReason = fmt.Sprintf(
				"Error getting process for incoming envelope %s: %s", envelope.ID, err)
			s.Log.Errorf(statusReason)
		} else {
			s.Log.Debugf("Process ID: %s", process.ID)
			sourceRef = process.SourceRef
		}
	}

	err = s.Conn.Publish(reply, api.EnvelopeAck{
		Response:        api.Response{Status: api.StatusOK},
		ID:              envelope.ID,
		ReceptionStatus: receptionStatus,
		Reason:          statusReason,
	})

	if err != nil {
		s.Log.Errorf("Error replying: %s", err)
	}

	if process != nil {
		var storageReceptionStatus storage.ReceptionStatus
		switch receptionStatus {
		case api.ReceptionReceiving:
			storageReceptionStatus = storage.ReceptionOngoing
		case api.ReceptionAccepted:
			storageReceptionStatus = storage.ReceptionDone
		}
		if err := process.UpdateReceiveStatus(
			envelope.ID, actor.ID, sourceRef, storageReceptionStatus,
		); err != nil {
			s.Log.Errorf("Error updating receive status: %s", err)
			return
		}

		targets, err := process.GetTargets(sourceRef, envelope.ID, false)
		if err != nil {
			s.Log.Errorf("Error getting targets: %s", err)
			return
		}

		s.Log.Noticef("Will forward to %s", targets)

		for _, target := range targets {
			s.emitEnvelope(process, envelope.ID, target)
		}
	}
}

func (s *Service) emitEnvelope(process *process.Process, envelopeID uuid.UUID, target api.EnvelopeTarget) {
	s.Conn.Publish(api.EmissionSignalSubjet, api.EmissionSignal{
		ProcessID:  process.ID,
		EnvelopeID: envelopeID,
		Target: api.EnvelopeTarget{
			ActorID: target.ActorID,
			NodeID:  target.NodeID,
			Input:   target.Input,
		},
	})
}
