// Copyright 2016 Orus.io. All rights reserved.

package emission

import (
	"fmt"
	"runtime/debug"
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/nats-io/go-nats"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
)

const (
	// EmissionQueue is the name we use to queueSub to a queue with the emission
	EmissionQueue = "emission_queue"
)

// Options is used to create a Emission
type Options struct {
	services.Options

	ReceptionTimeout time.Duration

	ActorStorage    storage.ActorStorage
	EnvelopeStorage storage.EnvelopeStorage
	ProcessManager  *process.Manager
}

// NewService creates a Emission
func (o Options) NewService() *Service {
	if o.ReceptionTimeout == 0 {
		o.ReceptionTimeout = 2 * time.Second
	}

	return &Service{
		BaseService:  services.NewBaseService("emission", o.Options),
		o:            o,
		quit:         make(chan struct{}),
		emitRoutines: make(map[api.EmissionSignal]chan struct{}),
	}
}

// Service is in charge of emitting envelopes to Actors
type Service struct {
	services.BaseService

	o Options

	quit         chan struct{}
	emitRoutines map[api.EmissionSignal]chan struct{}
}

// Startup starts the service
func (s *Service) Startup() error {
	s.Log.Noticef("Startup")

	if err := s.SubscribeAll(
		services.QueueSubscribe(api.EmissionSignalSubjet, EmissionQueue, s.handleSignal),
		services.QueueSubscribe(api.CoreActorEventsSubject, EmissionQueue, s.handleActorEvent),
		services.QueueSubscribe(api.CoreProcessEventsSubject, EmissionQueue, s.handleProcessEvent),
	); err != nil {
		s.Shutdown()
		return err
	}

	// if e.isAlone(), reschedule pending emissions

	return nil
}

// Shutdown signal the routines to stop and unsubscribe
func (s *Service) Shutdown() error {
	err := s.UnsubscribeAll()
	close(s.quit)
	s.Tasks.Wait()
	return err
}

func (s *Service) handleActorEvent(subject string, event api.CoreEvent) {
	s.TaskAdd()
	defer s.SafeTaskDone()

	if event.Name == "up" {
		// if "up", reschedule emissions
		s.Log.Noticef("%s is back online", event.Actor.ID)
		emissions := s.o.ProcessManager.ListActiveEmissions(event.Actor.ID)

		for _, pending := range emissions {
			s.Log.Noticef("Rescheduling %s to %s", pending.EnvelopeID, pending.TargetActor)
			s.Conn.Publish(api.EmissionSignalSubjet, api.EmissionSignal{
				ProcessID:  pending.ProcessID,
				EnvelopeID: pending.EnvelopeID,
				Target: api.EnvelopeTarget{
					ActorID: pending.TargetActor,
					NodeID:  pending.TargetInputRef.NodeID,
					Input:   pending.TargetInputRef.Input,
				},
			})
		}
	}
}

func (s *Service) handleProcessEvent(subject string, event api.CoreEvent) {
	s.TaskAdd()
	defer s.SafeTaskDone()

	if event.Name == "status" && event.Process.Status == api.ProcessRunning {
		s.Log.Noticef("Process %s is now running")

		emissions := s.o.ProcessManager.ListProcessEmissions(event.Process.ID)

		for _, pending := range emissions {
			s.Log.Noticef("Rescheduling %s to %s", pending.EnvelopeID, pending.TargetActor)
			s.Conn.Publish(api.EmissionSignalSubjet, api.EmissionSignal{
				ProcessID:  pending.ProcessID,
				EnvelopeID: pending.EnvelopeID,
				Target: api.EnvelopeTarget{
					ActorID: pending.TargetActor,
					NodeID:  pending.TargetInputRef.NodeID,
					Input:   pending.TargetInputRef.Input,
				},
			})
		}
	}
}

func trySignal(c chan<- struct{}) (success bool) {
	defer func() {
		recover()
	}()
	c <- struct{}{}
	return true
}

func (s *Service) handleSignal(subject string, signal api.EmissionSignal) {
	s.TaskAdd()
	defer s.SafeTaskDone()

	c, ok := s.emitRoutines[signal]
	if ok && trySignal(c) {
		return
	}
	c = make(chan struct{})
	s.emitRoutines[signal] = c

	process, err := s.o.ProcessManager.Get(signal.ProcessID)
	if err != nil {
		s.Log.Errorf("Received a wrong processID: %s", err)
		return
	}

	s.TaskAdd()
	go func() {
		s.emit(c, process, signal.EnvelopeID, signal.Target)
		close(c)
		delete(s.emitRoutines, signal)
	}()
}

// emit a fragment of envelope to the target.
func (s *Service) emitFragment(process *process.Process, envelopeID uuid.UUID, target api.EnvelopeTarget, actor api.Actor) (sent bool, complete bool, err error) {
	status := process.GetStatus()
	if status != api.ProcessRunning {
		s.Log.Noticef("Process not running (current state: %s), skip emission", status)
		return false, true, nil
	}

	emitPosition, err := process.GetEmitPosition(envelopeID, target)
	if err != nil {
		return false, false, fmt.Errorf("Cannot get emit position: %s", err)
	}
	defer emitPosition.Unlock()

	pos, err := emitPosition.Get()
	if err != nil {
		return false, false, fmt.Errorf("Cannot read emit position: %s", err)
	}

	if pos.Complete {
		s.Log.Debugf("Emission already complete, skipping")
		return false, true, nil
	}

	newenv, newpos, err := s.o.EnvelopeStorage.ReadEnvelope(envelopeID, pos, api.EnvelopeMaxSize)
	if err != nil {
		return false, false, fmt.Errorf("Error reading envelope for transfer: %s", err)
	}
	if newpos.Equals(pos) {
		s.Log.Debugf("No new data to emit")
		return false, false, nil
	}

	newenv.Context = api.EnvelopeContext{
		ProcessID: process.ID,
		NodeID:    target.NodeID,
	}

	var ack api.EnvelopeAck
	outsubject := api.ActorInputSubject(target.ActorID, target.Input)
	// For the non-first fragments of envelope, publish to a dedicated subject
	if !pos.Start {
		outsubject += "." + envelopeID.String()
	}
	if actor.AccountID.IsZero() {
		err = s.Conn.Request(outsubject, newenv, &ack, s.o.ReceptionTimeout)
	} else {
		err = s.Conn.RequestFor(actor.AccountID.String(), outsubject, newenv, &ack, s.o.ReceptionTimeout)
	}
	if err == nats.ErrTimeout {
		err := fmt.Errorf(
			"Actor '%s' did not ack the envelope on %s. "+
				"Mark it unresponsive.",
			actor.Name, outsubject)
		if err := s.o.ActorStorage.MarkUnresponsive(target.ActorID); err != nil {
			panic(fmt.Sprintf("Cannot mark the actor unresponsive: %s", err))
		}
		return false, false, err
	} else if err != nil {
		return false, false, fmt.Errorf(
			"Error transmitting envelope to %s on %s: %s",
			target, outsubject, err)
	}
	if ack.ReceptionStatus == api.ReceptionError {
		return false, false, fmt.Errorf("Actor returned an error: %s", ack.Reason)
	}
	s.Log.Noticef("Published to %s", outsubject)
	if err := emitPosition.Set(newpos); err != nil {
		return false, false, fmt.Errorf("Cannot set emit position: %s", err)
	}
	return true, pos.Complete, nil
}

func (s *Service) emit(signal <-chan struct{}, process *process.Process, envelopeID uuid.UUID, target api.EnvelopeTarget) {
	s.Log.Noticef("Sending to %s", target)
	defer s.SafeTaskDone()

	defer func() {
		if r := recover(); r != nil {
			s.Log.Noticef("Error in emit: %s\n%s", r, debug.Stack())
		}
	}()

	actor, ok := s.o.ActorStorage.Get(target.ActorID)
	if !ok {
		s.Log.Errorf("Unknown actor")
		return
	}

	timer := time.NewTimer(time.Second)

	for {
		timer.Stop()
		// Emit as much of the envelope we can
		for {
			sent, complete, err := s.emitFragment(process, envelopeID, target, actor)
			if err != nil {
				s.Log.Errorf("%s", err)
				return
			}
			if complete {
				return
			}
			if !sent {
				break
			}
		}
		// After 10 minutes, stop the routine. It will be recreated when a new
		// fragment has to be sent, or never if the process got stopped
		timer.Reset(10 * time.Minute)
		// wait for a signal
		select {
		case <-signal:
			continue
		case <-s.quit:
			timer.Stop()
			return
		case <-timer.C:
			return
		}
	}
}
