// Copyright 2016 Orus.io. All rights reserved.

package registration

import (
	"encoding/pem"
	"github.com/m4rw3r/uuid"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
)

// Options of the Service
type Options struct {
	services.Options

	AccountStorage storage.AccountStorage
	ActorStorage   storage.ActorStorage
	ServerCA       []byte
}

// NewService instanciate a Service
func NewService(o Options) *Service {
	return &Service{
		o:           o,
		BaseService: services.NewBaseService("registration", o.Options),
	}
}

// Service implements the registration api
type Service struct {
	services.BaseService
	o Options
}

// Startup subscribes the service to the registration apis
func (s *Service) Startup() error {
	if err := s.SubscribeAll(
		services.Subscribe(api.Registration, s.handleRegistration),
		services.Subscribe(api.Registration+".*", s.handleRegistration),
	); err != nil {
		s.Shutdown()
		return err
	}
	return nil
}

func parseSubject(subject string) (anon bool, accountID uuid.UUID) {
	if subject == api.Registration {
		anon = true
		return
	}
	if len(subject) != len(api.Registration)+37 {
		panic("Invalid registration subject: " + subject)
	}
	strID := subject[len(api.Registration):]
	return false, uuid.MustFromString(strID)
}

func (s *Service) handleRegistration(subject, reply string, req api.RegistrationReq) {
	var rep api.RegistrationRep
	s.TaskAdd()
	defer s.SendReply(reply, &rep)

	// see if subject contains the account ID. If so, it is an existing account
	// and the message source is guaranted.
	anon, accountID := parseSubject(subject)

	var account api.Account

	// If non-existing account, create it in the adequate state and its csr attached
	if anon {
		var (
			ok  bool
			err error
		)
		account, ok = s.o.AccountStorage.GetByName(req.Name)
		if !ok {
			account, err = s.o.AccountStorage.Create(api.Account{
				Name:   req.Name,
				Type:   req.Type,
				Status: api.AccountStatusPending,
				CSR:    req.CSR,
			})
			if err != nil {
				rep.Response = api.ServerError(err)
				return
			}
		}
	} else {
		var exists bool
		account, exists = s.o.AccountStorage.Get(accountID)
		if !exists {
			rep.Response = api.ClientError("No such account %s", exists)
			return
		}
	}

	// For each actor, if not-yet created (ie if this is an original req),
	// create the actor in the pending state
	for _, actor := range req.ActorList {
		if actor.ID.IsZero() {
			sActor := api.Actor{
				Name:      actor.Name,
				Kind:      actor.Kind,
				AccountID: account.ID,
				Status:    api.ActorPending,
				Roles:     actor.Roles,
			}
			var err error
			sActor, err = s.o.ActorStorage.Create(sActor)
			if err != nil {
				rep.Response = api.ClientError(
					"Error creating Actor '%s': %s",
					actor.Name, err)
				return
			}
		}
	}

	// Check the current state of the account and actors and put it all
	// in the reply, along with the signed certificate if applicable
	rep.Status = api.StatusOK
	rep.ID = account.ID
	switch account.Status {
	case api.AccountStatusPending:
		rep.RegistrationStatus = api.RegistrationPending
	case api.AccountStatusEnabled:
		rep.RegistrationStatus = api.RegistrationAccepted
	case api.AccountStatusDisabled:
		if account.Cert == "" {
			rep.RegistrationStatus = api.RegistrationRejected
		} else {
			rep.RegistrationStatus = api.RegistrationAccepted
		}
	}
	if rep.RegistrationStatus == api.RegistrationAccepted {
		rep.SignedCertificate = account.Cert
		if len(s.o.ServerCA) != 0 {
			rep.ServerCA = string(pem.EncodeToMemory(&pem.Block{
				Type:  "CERTIFICATE",
				Bytes: s.o.ServerCA,
			}))
		}
	}

	actorList := s.o.ActorStorage.ListAccount(account.ID)

	for _, actor := range actorList {
		rep.ActorList = append(rep.ActorList, api.Actor{
			ID:     actor.ID,
			Name:   actor.Name,
			Kind:   actor.Kind,
			Status: actor.Status,
			Roles:  actor.Roles,
		})
	}
}
