package storage

import (
	"github.com/m4rw3r/uuid"
)

// NodeType is the type of Node
//go:generate enum-marshaler -type NodeType -suffix Node -lower graph.go
type NodeType int

const (
	// EmitterNode is the type of emitter nodes
	EmitterNode NodeType = iota
	// ConsumerNode is the type of consumer nodes
	ConsumerNode
	// WorkerNode is the type of worker nodes
	WorkerNode
)

// GraphStatus is the status of the graph
//go:generate enum-marshaler -type GraphStatus -prefix Graph -lower graph.go
type GraphStatus int

const (
	// GraphDraft is for graphs that still can be modified
	GraphDraft GraphStatus = iota
	// GraphPublished is for graphs that cannot be modified anymore
	GraphPublished
	// GraphActive is for graphs that can be triggered by incoming envelopes
	GraphActive
)

// SourceMatch are rules to match incoming envelope with
type SourceMatch struct {
	EventTypes []string // All event types envelope must contain exactly
}

// Node is a graph vertex
type Node struct {
	ID            string
	Type          NodeType
	ActorIDs      []uuid.UUID // Explicit actors. Deprecated (will be removed in a later version)
	Actors        []string    // Explicit actors, by name or UUID
	Roles         []string    // Actor roles
	RoleBroadcast bool        // If true, all actors matching each role will receive the message, instead of only one actor per role

	SourceMatch *SourceMatch // define only if Type==EmitterNode

	Inputs  []string
	Outputs []string

	// TODO add 'arguments'
}

// NodeInputRef is a reference to a node input
type NodeInputRef struct {
	NodeID string
	Input  string
}

// NodeOutputRef is a reference to a node input
type NodeOutputRef struct {
	NodeID string
	Output string
}

// Edge is a Graph oriented edge
type Edge struct {
	From NodeOutputRef
	To   NodeInputRef
}

// Graph describe a processing graph.
type Graph struct {
	ID      uuid.UUID
	Name    string
	Version string
	Status  GraphStatus
	Nodes   []Node
	Edges   []Edge
}

// GraphStorage handle Graph persistence
type GraphStorage interface {
	Load(uuid.UUID) (Graph, bool)
	Save(Graph) (uuid.UUID, error)
	LoadActiveGraph() []Graph
	Query(Graph) []Graph
}
