package storage

import (
	"errors"
	"time"

	"github.com/m4rw3r/uuid"

	"bitbucket.org/orus-io/go-xbus/api"
)

// ProcessEventStatusHandler is the type of function that handle
// status changes
type ProcessEventStatusHandler func(p Process, old api.ProcessStatus, new api.ProcessStatus)

// NoOpProcessEventStatusHandler is a noop handler
func NoOpProcessEventStatusHandler(Process, api.ProcessStatus, api.ProcessStatus) {}

// Process contains the data necessary for process persistence
type Process struct {
	ID                uuid.UUID
	TriggerEmitterID  uuid.UUID
	TriggerEnvelopeID uuid.UUID
	GraphID           uuid.UUID
	SourceOutput      NodeOutputRef
}

// EmitPosition provides exclusive access to a emit position
type EmitPosition interface {
	Get() (EnvelopeReadPos, error)
	Set(EnvelopeReadPos) error
	Unlock() error
}

// Emission identifies the emission of an envelope to a target
type Emission struct {
	ProcessID      uuid.UUID
	EnvelopeID     uuid.UUID
	TargetActor    uuid.UUID
	TargetInputRef NodeInputRef
}

// Reception identifies a reception of an envelop from an actor
type Reception struct {
	ProcessID       uuid.UUID
	EnvelopeID      uuid.UUID
	SourceActor     uuid.UUID
	SourceOutputRef NodeOutputRef
}

// ReceptionStatus describes the status of an envelope reception in
// the context of a process
type ReceptionStatus int

const (
	// ReceptionPending is when envelop reception is not yet started
	ReceptionPending ReceptionStatus = iota
	// ReceptionOngoing is when part(s) of the envelope were received, but not all
	ReceptionOngoing
	// ReceptionDone is when all the complete envelop has been receiced
	ReceptionDone
	// ReceptionError is when a reception error occured
	ReceptionError
	// ReceptionClosed is when no envelope is to be expected
	ReceptionClosed
)

// ProcessNodeEnd signals the end of processing by a process node
type ProcessNodeEnd struct {
	Node       ProcessNode
	EnvelopeID uuid.UUID
	Result     api.ActorProcessingResultType
}

// ProcessLogEntry is the unique source of information for process tracking.
type ProcessLogEntry struct {
	Timestamp        time.Time
	Emission         *Emission        `json:",omitempty"`
	EmissionPosition *EnvelopeReadPos `json:",omitempty"`
	Reception        *Reception       `json:",omitempty"`
	ReceptionStatus  ReceptionStatus  `json:",omitempty"`
	ProcessNodeEnd   *ProcessNodeEnd  `json:",omitempty"`
}

// A ProcessNode is a live graph node. One graph node can correspond to several
// ProcessNode (especially consumer nodes)
type ProcessNode struct {
	ActorID uuid.UUID
	NodeID  string
}

// ProcessNodeStatus is the process node status
//go:generate enum-marshaler -type ProcessNodeStatus -prefix ProcessNode -lower -tests
type ProcessNodeStatus int

const (
	// ProcessNodeStatusPending is before node activation
	ProcessNodeStatusPending ProcessNodeStatus = iota
	// ProcessNodeActive is when any input or output is active
	ProcessNodeActive
	// ProcessNodeDone is when all input/output are done transmitting
	ProcessNodeDone
	// ProcessNodeUnreachable is when every path leading to the node
	// are closed and the node will never be active
	ProcessNodeUnreachable
	// ProcessNodeError is when an error occured on the node
	ProcessNodeError
)

// ProcessNodeIOStatus reflects the status of the inputs or outputs of a node
//go:generate enum-marshaler -type ProcessNodeIOStatus -prefix ProcessNodeIO -lower -tests
type ProcessNodeIOStatus int

const (
	// ProcessNodeIOPending is when no io started yet
	ProcessNodeIOPending ProcessNodeIOStatus = iota
	// ProcessNodeIOStarted is when one or more io started emitting
	ProcessNodeIOStarted
	// ProcessNodeIOAllStarted is when all io started emitting
	ProcessNodeIOAllStarted
	// ProcessNodeIOClosed is when all the io is closed without any transmission
	ProcessNodeIOClosed
	// ProcessNodeIODone is when all io finished emitting
	ProcessNodeIODone
	// ProcessNodeIOError is when an error occured on one of the io
	ProcessNodeIOError
)

// ProcessNodeIOStatusMap is a map of emission or reception statuses
type ProcessNodeIOStatusMap map[string]ProcessNodeIOStatus

// ProcessNodeState reflect the current state of a process node.
type ProcessNodeState struct {
	Status         ProcessNodeStatus
	Input          ProcessNodeIOStatus
	InputStatuses  ProcessNodeIOStatusMap
	Output         ProcessNodeIOStatus
	OutputStatuses ProcessNodeIOStatusMap
}

// ProcessNodeStateMap contains states of process nodes
type ProcessNodeStateMap map[ProcessNode]ProcessNodeState

// ProcessState reflect the current state of a process execution.
// It is completely updated by applying incoming logentries
type ProcessState struct {
	Status     api.ProcessStatus
	NodeStates ProcessNodeStateMap
}

// NextStateCallback is a callback that calculate the next state of a process
type NextStateCallback func(
	process Process, state ProcessState, logentry ProcessLogEntry,
) (ProcessState, error)

// UndefinedNextStateCallback is NextStateCallback that always return
// a 'undefined' error
func UndefinedNextStateCallback(
	process Process, state ProcessState, logentry ProcessLogEntry,
) (ProcessState, error) {
	return ProcessState{}, errors.New("NextStateCallback undefined")
}

// ProcessStorage persists process data
type ProcessStorage interface {
	GetProcess(uuid.UUID) (Process, bool)
	GetProcessByTrigger(emitter api.Actor, envelopeID uuid.UUID) (Process, bool)

	Query(includeTerminated bool) []Process

	CreateProcess(Process) (uuid.UUID, error)

	// GetTargets retrieve targets for an envelope from a routing
	// table. If the table has no matching entry, a callback is used to
	// determine the targets, and the table is updated.

	// When the table is updated, ProcessLogEntry of type 'emission' are
	// generated and the process state is updated accordingly.

	// In some cases we do not want the table to be updated (more especially we do
	// not want log entries to be generated), so the function has a
	// 'noRouteTableUpdate' argument.

	GetTargets(
		processID uuid.UUID,
		source NodeOutputRef,
		envelopeID uuid.UUID,
		noRouteTableUpdate bool,
		calcTargets func() ([]api.EnvelopeTarget, error),
	) ([]api.EnvelopeTarget, error)

	UpdateReceiveStatus(reception Reception, status ReceptionStatus) error

	// GetEmitPosition returns an exclusive access to a emit position.
	// It _must_ be unlocked after use
	GetEmitPosition(Emission) (EmitPosition, error)

	ProcessNodeEnd(
		process uuid.UUID, node ProcessNode,
		envelopeID uuid.UUID, result api.ActorProcessingResultType) error

	// All functions listing emissions return extra unstarted emissions
	// for each route that has no emission yet.
	ListActiveEmissions(targetActor uuid.UUID) []Emission
	ListProcessEmissions(process uuid.UUID) []Emission

	GetState(uuid.UUID) ProcessState
	SetStatus(uuid.UUID, api.ProcessStatus) error

	SetNextStateCallback(NextStateCallback)
	SetEventHandlers(onStatusChanged ProcessEventStatusHandler)
}
