// Copyright 2016 Orus.io. All rights reserved.

package sql

import (
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/slicebit/qb"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
)

// NewActorStorage initialize a new ActorStorage
func NewActorStorage(model *model.Model, db *yago.DB, logger xbus.Logger) *ActorStorage {
	return &ActorStorage{
		Base: Base{
			model:  model,
			db:     db,
			logger: logger,
		},
		onActorUp:   storage.NoopActorEventHandler,
		onActorDown: storage.NoopActorEventHandler,
	}
}

// ActorStorage implements all the persistor interfaces
type ActorStorage struct {
	Base

	onActorUp   storage.ActorEventHandler
	onActorDown storage.ActorEventHandler
}

func actorModelToAPI(actor model.Actor) api.Actor {
	return api.Actor{
		ID:        actor.ID,
		Name:      actor.Name,
		AccountID: actor.AccountID.UUID,
		Kind:      actor.Kind,
		Status:    actor.Status,
		Roles:     []string{actor.Role},
	}
}

func actorAPIToModel(actor api.Actor) model.Actor {
	return model.Actor{
		Base:   model.Base{ID: actor.ID},
		Name:   actor.Name,
		Kind:   actor.Kind,
		Role:   actor.Roles[0],
		Status: actor.Status,
		AccountID: uuid.NullUUID{
			Valid: !actor.AccountID.IsZero(), UUID: actor.AccountID},
	}
}

func actorListModelToAPI(actors []model.Actor) (out []api.Actor) {
	for _, actor := range actors {
		out = append(out, actorModelToAPI(actor))
	}
	return
}

// SetEventHandlers sets the event handlers
func (p *ActorStorage) SetEventHandlers(onActorUp, onActorDown storage.ActorEventHandler) {
	if onActorUp == nil {
		p.onActorUp = storage.NoopActorEventHandler
	} else {
		p.onActorUp = onActorUp
	}
	if onActorDown == nil {
		p.onActorDown = storage.NoopActorEventHandler
	} else {
		p.onActorDown = onActorDown
	}
}

// Get returns an actor from its id
func (p ActorStorage) Get(id uuid.UUID) (api.Actor, bool) {
	var (
		actor model.Actor
	)
	tx := p.MustBegin()
	defer tx.Rollback()

	err := tx.Query(p.model.Actor).Get(&actor, id)
	if err == yago.ErrRecordNotFound {
		return api.Actor{}, false
	}
	if err != nil {
		panic(err)
	}
	return actorModelToAPI(actor), true
}

// GetByName returns an actor from its name
func (p ActorStorage) GetByName(name string) (api.Actor, bool) {
	var (
		actor model.Actor
	)
	tx := p.MustBegin()
	defer tx.Rollback()
	err := tx.Query(p.model.Actor).Where(p.model.Actor.Name.Eq(name)).One(&actor)
	if err == yago.ErrRecordNotFound {
		return api.Actor{}, false
	}
	if err != nil {
		panic(err)
	}
	return actorModelToAPI(actor), true
}

func (p ActorStorage) list(clauses ...qb.Clause) []api.Actor {
	var actors []model.Actor

	tx := p.MustBegin()
	defer tx.Rollback()

	q := tx.Query(p.model.Actor)
	if len(clauses) != 0 {
		q = q.Where(clauses...)
	}

	must(q.All(&actors))

	return actorListModelToAPI(actors)
}

// List returns all the actors
func (p ActorStorage) List() []api.Actor {
	return p.list()
}

// ListAccount returns a list of all actors the specified account provides access to.
func (p ActorStorage) ListAccount(accountID uuid.UUID) []api.Actor {
	return p.list(
		p.model.Actor.AccountID.Eq(accountID),
	)
}

// ListKind returns a list of all actors of one type
func (p ActorStorage) ListKind(actorKind api.ActorKind) []api.Actor {
	return p.list(
		p.model.Actor.Kind.Eq(actorKind),
	)
}

// ListRole returns a list of all actors of one role
func (p ActorStorage) ListRole(actorKind api.ActorKind, role string) []api.Actor {
	return p.list(
		p.model.Actor.Kind.Eq(actorKind),
		p.model.Actor.Role.Eq(role),
	)
}

// Create creates an Actor
func (p ActorStorage) Create(actor api.Actor) (api.Actor, error) {
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	dbActor := actorAPIToModel(actor)
	err := tx.Insert(&dbActor)
	if err != nil {
		// TODO identify constraint errors and return them (not panic)
		panic(err)
	}
	tx.MustCommit()
	return actorModelToAPI(dbActor), nil
}

// Update updates an Actor
func (p ActorStorage) Update(actor api.Actor) (api.Actor, error) {
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	dbActor := actorAPIToModel(actor)
	err := tx.Update(&dbActor)
	// TODO identify constraint errors and return them
	if err != nil {
		panic(err)
	}
	tx.MustCommit()
	p.logger.Noticef("Updated %+v", dbActor)
	return actorModelToAPI(dbActor), nil
}

// JustSeen update the actor lastSeen timestamp
func (p ActorStorage) JustSeen(id uuid.UUID) error {
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	var actor model.Actor
	err := tx.Query(p.model.Actor).Get(&actor, id)
	if err == yago.ErrRecordNotFound {
		return err
	}
	if err != nil {
		panic(err)
	}
	upEvent := !actor.Visible
	actor.Visible = true
	actor.LastSeen = time.Now()
	err = tx.Update(&actor, model.ActorVisible, model.ActorLastSeen)
	if err != nil {
		// TODO if constraint error, return a user friendly error
		panic(err)
	}
	tx.MustCommit()
	if upEvent {
		p.onActorUp(actorModelToAPI(actor))
	}
	return nil
}

// MarkUnresponsive marks the actor as unresponsive
func (p ActorStorage) MarkUnresponsive(id uuid.UUID) error {
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	var actor model.Actor
	err := tx.Query(p.model.Actor).Get(&actor, id)
	if err == yago.ErrRecordNotFound {
		return err
	}
	if err != nil {
		panic(err)
	}
	downEvent := actor.Visible
	if actor.Visible {
		actor.Visible = false
		err = tx.Update(&actor, model.ActorVisible)
		if err != nil {
			// TODO if constraint error, return a user friendly error
			panic(err)
		}
	}
	tx.MustCommit()
	if downEvent {
		p.onActorDown(actorModelToAPI(actor))
	}
	return nil
}
