// Copyright 2016 Orus.io. All rights reserved.

// Tests around actors.

package sql

import (
	"testing"

	"github.com/m4rw3r/uuid"
	"github.com/stretchr/testify/assert"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
	"bitbucket.org/orus-io/xbus/testutils"
	"bitbucket.org/orus-io/xbus/testutils/sqldb"
)

// Tests around the link between actors and accounts.
func TestActorAccountLink(t *testing.T) {

	// Init.
	logger := testutils.GetTestLogger(t)
	defer logger.VerifyExpectations()
	m, db, closeDB := sqldb.GetTestDB(t)
	defer closeDB()
	actorStorage := NewActorStorage(m, db, logger)

	// Generate a random name.
	uuidobj, _ := uuid.V4()
	name := uuidobj.String()

	// Create an account.
	account := model.Account{
		Name: name,
	}
	if err := db.Insert(&account); err != nil {
		t.Fatal(err)
	}

	// Create an actor.
	actor := model.Actor{
		Name: name,
		Kind: api.EmitterKind,
	}
	if err := db.Insert(&actor); err != nil {
		t.Fatal(err)
	}

	// The actor -> account link should be null at this point.
	if actorInStorage, exists := actorStorage.Get(actor.ID); !exists {
		t.Fatal("actor missing")
	} else {
		assert.True(t, actorInStorage.AccountID.IsZero())
	}

	// The "ListAccount" function should not find the actor so far.
	assert.Empty(t, actorStorage.ListAccount(account.ID))

	// Set a link.
	actor.AccountID.Scan(account.ID.String())
	if err := db.Update(&actor, model.ActorAccountID); err != nil {
		t.Fatal(err)
	}

	// The actor -> account link should now exist.
	if actorInStorage, exists := actorStorage.Get(actor.ID); !exists {
		t.Fatal("actor missing")
	} else {
		assert.False(t, actorInStorage.AccountID.IsZero())
	}

	// The "ListAccount" function should now find the actor.
	actorsInStorage := actorStorage.ListAccount(account.ID)
	assert.Len(t, actorsInStorage, 1)
	assert.Equal(t, account.ID, actorsInStorage[0].AccountID)

	// Ensure the account can be deleted - ONDELETE SET NULL behavior.
	if err := db.Delete(&account); err != nil {
		t.Fatal(err)
	}
	if actorInStorage, exists := actorStorage.Get(actor.ID); !exists {
		t.Fatal("actor missing")
	} else {
		assert.True(t, actorInStorage.AccountID.IsZero())
	}
}
