// Copyright 2016 Orus.io. All rights reserved.

package sql

import (
	"errors"
	"fmt"

	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/tools"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
)

// NewEnvelopeStorage initialize a EnvelopeStorage
func NewEnvelopeStorage(model *model.Model, db *yago.DB, logger xbus.Logger) EnvelopeStorage {
	return EnvelopeStorage{Base{
		model:  model,
		db:     db,
		logger: xbus.NewPrefixedLogger("[storage/sql/envelop]", logger),
	}}
}

// EnvelopeStorage implements the storage.EnvelopeStorage interface
type EnvelopeStorage struct {
	Base
}

// GetEventTypes returns the eventtypes of an envelope, or nil if unknown
// exists is false if the envelope is unknown, true if exists.
// will panic on any SQL error.
func (es EnvelopeStorage) GetEventTypes(id uuid.UUID) (types []string, exists bool) {
	tx := es.MustBegin()
	defer tx.Rollback()

	var (
		dbEnvelope model.Envelope
	)

	err := tx.Query(es.model.Envelope).Get(&dbEnvelope, id)
	if err == yago.ErrRecordNotFound {
		return nil, false
	}
	if err != nil {
		panic(err)
	}
	if !dbEnvelope.TypesKnown {
		return nil, true
	}
	rows, err := tx.Query(es.model.Event).
		Select(es.model.Event.EventType).
		Where(es.model.Event.EnvelopeID.Eq(id)).
		SQLQuery()
	if err != nil {
		panic(err)
	}
	var eventTypes []string

	for rows.Next() {
		var eventType string
		must(
			rows.Scan(&eventType))
		eventTypes = append(eventTypes, eventType)
	}
	return eventTypes, true
}

// StoreEnvelope stores an envelope chunk and returns the envelope status in the
// storage
func (es EnvelopeStorage) StoreEnvelope(envelope api.Envelope) (storage.EnvelopeStatus, error) {
	var (
		dbEnvelope model.Envelope
		status     storage.EnvelopeStatus
	)

	tx := es.MustBegin()
	defer tx.RollbackIfOpened()
	// Load the envelope or create it if it is a new one
	err := tx.Query(es.model.Envelope).Get(&dbEnvelope, envelope.ID)
	if err == yago.ErrRecordNotFound {
		dbEnvelope = model.Envelope{
			Base:       model.Base{ID: envelope.ID},
			TypesKnown: false,
			Status:     "receiving",
		}
		err = tx.Insert(&dbEnvelope)
	}
	if err != nil {
		panic(err)
	}

	// Load the current status of the envelope

	status.ID = envelope.ID
	status.TypesKnown = dbEnvelope.TypesKnown

	switch dbEnvelope.Status {
	case "complete":
		err = fmt.Errorf("Envelope '%s' is already complete", envelope.ID)
		es.logger.Errorf(err.Error())
		return storage.EnvelopeStatus{}, err
	case "error":
		status.Status = "error"
		status.Err = errors.New(dbEnvelope.Error)
		return status, nil
	case "receiving":
	default:
		return storage.EnvelopeStatus{},
			fmt.Errorf("Invalid status: %s", dbEnvelope.Status)
	}

	// Load the Event & their validators

	var (
		dbEvents []model.Event
	)

	must(
		tx.Query(es.model.Event).
			Where(es.model.Event.EnvelopeID.Eq(dbEnvelope.ID)).
			All(&dbEvents))

	var (
		dbEventByID     = make(map[uuid.UUID]model.Event)
		eventValidators = make(map[uuid.UUID]tools.EventValidator)
		eventStatuses   = make(map[uuid.UUID]storage.EventStatus)
	)

	for _, dbEvent := range dbEvents {
		availableItems := 0
		if dbEvent.Validator != "" {
			validator, err := tools.LoadEventValidator(dbEvent.Validator)
			if err != nil {
				panic(err)
			}
			eventValidators[dbEvent.EventID] = validator
			availableItems = validator.CurrentItemCount
		}
		dbEventByID[dbEvent.EventID] = dbEvent
		eventStatuses[dbEvent.EventID] = storage.EventStatus{
			ID:             dbEvent.EventID,
			Type:           dbEvent.EventType,
			AvailableItems: availableItems,
			TotalItems:     dbEvent.ItemCount,
			Status:         dbEvent.Status,
		}
	}

	// Now load the incoming events into the database.
	// If the batch n° is consistent with the validator current
	// state, we feed it immediately to it, followed by any
	// waiting batch if their batch n° match.
	// If not, only store the items to the db.

	for _, event := range envelope.Events {
		dbEvent, ok := dbEventByID[event.ID]
		if !ok {
			dbEvent = model.Event{
				EventID:    event.ID,
				EnvelopeID: dbEnvelope.ID,
				Status:     "receiving",
				ItemCount:  event.ItemCount,
				Checksum:   event.Checksum,
			}
			if event.Type != "" {
				dbEvent.EventType = event.Type
			}
			must(
				tx.Insert(&dbEvent))
		}
		eventStatus, ok := eventStatuses[dbEvent.EventID]
		if !ok {
			eventStatus = storage.EventStatus{
				ID:             dbEvent.EventID,
				Type:           event.Type,
				AvailableItems: 0,
				TotalItems:     event.ItemCount,
				Status:         "receiving",
			}
		}

		// if no item in the event chunk, we need to write something
		// in the items table so batch numbers remain consistent
		// and ReadEnvelope can see empty batches and jump to non-empty ones
		if len(event.Items) == 0 {
			dbItem := model.Item{
				EventID: dbEvent.ID,
				Batch:   event.Index,
				Index:   -1,
			}
			must(
				tx.Insert(&dbItem))
		} else {
			// else store items in the database
			for i, item := range event.Items {
				dbItem := model.Item{
					EventID: dbEvent.ID,
					Batch:   event.Index,
					Index:   i,
					Data:    item,
				}
				must(
					tx.Insert(&dbItem))
			}
		}

		// if possible, update the validation state
		validator, _ := eventValidators[event.ID]
		if event.Index == validator.CurrentIndex+1 {
			validator, err = validator.Update(event)
			must(err)

			validatorStr, err := tools.DumpEventValidator(validator)
			must(err)

			dbEvent.Validator = validatorStr
			dbEvent.ItemCount = validator.ItemCount
			dbEvent.Checksum = validator.Checksum
			eventValidators[event.ID] = validator
			dbEventByID[event.ID] = dbEvent

			// TODO look for more items in the database and feed them
			// to the validator

			must(
				tx.Update(&dbEvent, "Validator", "ItemCount", "Checksum"))

			eventStatus.AvailableItems = validator.CurrentItemCount
			if validator.Complete {
				if validator.Valid {
					eventStatus.Status = "complete"
				} else if validator.Error != nil {
					eventStatus.Status = "error"
					eventStatus.Err = validator.Error
				}
			}
		}

		eventStatuses[dbEvent.EventID] = eventStatus
	}

	for _, eventStatus := range eventStatuses {
		status.Events = append(status.Events, eventStatus)
	}

	if !status.TypesKnown && len(envelope.EventIDs) != 0 {
		if len(envelope.EventIDs) == len(status.Events) {
			foundUnknownType := false
			for _, eventID := range envelope.EventIDs {
				status, _ := eventStatuses[eventID]
				if status.Type == "" {
					foundUnknownType = true
					break
				}
			}
			if !foundUnknownType {
				status.TypesKnown = true
				dbEnvelope.TypesKnown = true
				must(
					tx.Update(&dbEnvelope))
			}
		}
	}

	if status.TypesKnown {
		foundNotComplete := false
		foundErrors := false
		for _, eventStatus := range eventStatuses {
			if eventStatus.Status != "complete" {
				foundNotComplete = true
			}
			if eventStatus.Status == "error" {
				foundErrors = true
			}
		}
		if !foundNotComplete {
			status.Status = "complete"
		} else if foundErrors {
			status.Status = "error"
		} else {
			status.Status = "receiving"
		}
		dbEnvelope.Status = status.Status
		must(
			tx.Update(&dbEnvelope))
	}

	tx.MustCommit()

	return status, nil
}

// ReadEnvelope reads an envelope from the given position
func (es EnvelopeStorage) ReadEnvelope(id uuid.UUID, position storage.EnvelopeReadPos, maxSize int) (api.Envelope, storage.EnvelopeReadPos, error) {
	position = position.Clone()
	tx := es.MustBegin()
	defer tx.Rollback()

	var dbEnvelope model.Envelope
	// Load the envelope
	err := tx.Query(es.model.Envelope).Get(&dbEnvelope, id)
	if err == yago.ErrRecordNotFound {
		return api.Envelope{}, storage.EnvelopeReadPos{}, fmt.Errorf("No such envelope: %s", id)
	}
	must(err)

	if dbEnvelope.Status == "error" {
		return api.Envelope{}, position, fmt.Errorf("Envelope is in the error state: %s", dbEnvelope.Error)
	}

	// TODO if event types are not all known, return empty envelope
	// with unchanged position

	envelope := api.Envelope{
		ID: dbEnvelope.ID,
	}

	var dbEvents []model.Event

	must(
		tx.Query(es.model.Event).
			Where(es.model.Event.EnvelopeID.Eq(id)).
			All(&dbEvents))

	var estimateSize = 36 + len(dbEvents)*3

	allDone := true

	for _, dbEvent := range dbEvents {
		eventPos, _ := position.Events[dbEvent.EventID]
		eventPos.Index++

		selectItems := es.model.Item.GetMapper().Table().
			Select(es.model.Item.Batch, es.model.Item.Index, es.model.Item.Data).
			Where(
				es.model.Item.EventID.Eq(dbEvent.ID),
				es.model.Item.Index.NotEq(-1),
			).
			OrderBy(es.model.Item.Batch.Column, es.model.Item.Index.Column)

		if eventPos.ItemCount != 0 {
			offset := eventPos.ItemCount - 1

			selectItems = selectItems.Limit(offset, 9999999)
		}

		var emptyBatches []int
		{
			rows, err := es.db.Query(es.model.Item).
				Select(es.model.Item.Batch).
				Where(
					es.model.Item.EventID.Eq(dbEvent.ID),
					es.model.Item.Index.Eq(-1)).
				OrderBy(es.model.Item.Batch).
				SQLQuery()

			must(err)

			for rows.Next() {
				var batch int
				if err := rows.Scan(&batch); err != nil {
					rows.Close()
					panic(err)
				}
				emptyBatches = append(emptyBatches, batch)
			}

			rows.Close()
		}

		rows, err := tx.GetEngine().Query(selectItems)

		if err != nil {
			panic(err)
		}

		var (
			currentBatch int
		)

		if eventPos.ItemCount != 0 {
			if !rows.Next() {
				panic("Data inconsistency. The last item sent cannot be found.")
			}
			var i int
			must(rows.Scan(&currentBatch, &i, &api.Item{}))
			for len(emptyBatches) != 0 && emptyBatches[0] < currentBatch {
				emptyBatches = emptyBatches[1:]
			}
		}

		event := api.Event{
			ID:        dbEvent.EventID,
			Index:     eventPos.Index,
			Type:      dbEvent.EventType,
			ItemCount: dbEvent.ItemCount,
			Checksum:  dbEvent.Checksum,
		}

		envelope.EventIDs = append(envelope.EventIDs, event.ID)

		empty := true

		for rows.Next() {
			var (
				batch int
				index int
				item  []byte
			)
			if err := rows.Scan(&batch, &index, &item); err != nil {
				panic(err)
			}

			if batch-currentBatch > 1 {
				// There is a hole in the items list
				// Check for empty batches
				for len(emptyBatches) != 0 && emptyBatches[0] == currentBatch+1 {
					emptyBatches = emptyBatches[1:]
					currentBatch++
				}
				if batch-currentBatch > 1 {
					break
				}
			}

			currentBatch = batch

			if maxSize != 0 && estimateSize+len(item) >= maxSize {
				// Adding one more item exceed the maximum envelope size.
				// Stops here
				break
			}

			event.Items = append(event.Items, item)
			empty = false
			eventPos.ItemCount++
			estimateSize += len(item)
		}

		must(rows.Close())

		if eventPos.ItemCount != dbEvent.ItemCount {
			allDone = false
		}

		if !empty || position.Start || dbEnvelope.Status == "complete" {
			envelope.Events = append(envelope.Events, event)
			position.Events[event.ID] = eventPos
		}
	}

	position.Start = false
	position.Complete = dbEnvelope.Status == "complete" && allDone
	envelope.Last = position.Complete
	return envelope, position, nil
}
