// Copyright 2016 Orus.io. All rights reserved.

package storage

import (
	"testing"
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/stretchr/testify/assert"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
)

// EnvelopeStorageFactory is any factory of EnvelopeStorage
type EnvelopeStorageFactory func() (storage.EnvelopeStorage, func())

// EnvelopeStorageTests is a test suite for EnvelopeStorage implementations
func EnvelopeStorageTests(t *testing.T, factory EnvelopeStorageFactory, supportUnordered bool) {

	t.Run("Store/SingleEnvelope", func(t *testing.T) {
		envelopeStorage, clean := factory()
		defer clean()
		env := api.Envelope{
			ID: uuid.MustFromString("b74415ee-83c1-11e6-8f04-133b336287df"),
			Events: []api.Event{
				api.Event{
					ID:        uuid.MustFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
					Index:     1,
					Type:      "xbus.test",
					Items:     []api.Item{[]byte("Data")},
					ItemCount: 1,
					Checksum:  1859455373,
				},
			},
		}

		eventID := env.Events[0].ID
		env.EventIDs = []uuid.UUID{eventID}

		status, err := envelopeStorage.StoreEnvelope(env)
		assert.Nil(t, err)
		assert.Equal(t, env.ID, status.ID)
		assert.True(t, status.TypesKnown)
		assert.Equal(t, 1, len(status.Events))
		if t.Failed() {
			return
		}
		assert.Equal(t, eventID, status.Events[0].ID)
		assert.Equal(t, 1, status.Events[0].AvailableItems)
		assert.Equal(t, nil, status.Events[0].Err)
		assert.Equal(t, "complete", status.Events[0].Status)
		assert.Equal(t, "complete", status.Status)

		types, exists := envelopeStorage.GetEventTypes(env.ID)
		assert.True(t, exists)
		assert.Equal(t, []string{"xbus.test"}, types)
	})

	t.Run("Store/PairOfEnvelope", func(t *testing.T) {
		env1 := api.Envelope{
			ID: uuid.MustFromString("b74415ee-83c1-11e6-8f04-133b336287df"),
			Events: []api.Event{
				api.Event{
					ID:    uuid.MustFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
					Index: 1,
					Type:  "xbus.test",
					Items: []api.Item{[]byte("Data")},
				},
			},
		}

		eventID := env1.Events[0].ID
		env1.EventIDs = []uuid.UUID{eventID}

		env2 := api.Envelope{
			ID: env1.ID,
			Events: []api.Event{
				api.Event{
					ID:        eventID,
					Index:     2,
					Type:      "xbus.test",
					Items:     []api.Item{[]byte("Data")},
					ItemCount: 2,
					Checksum:  1307176161,
				},
			},
		}

		t.Run("Ordered", func(t *testing.T) {
			envelopeStorage, clean := factory()
			defer clean()

			status, err := envelopeStorage.StoreEnvelope(env1)
			assert.Nil(t, err)
			assert.Equal(t, env1.ID, status.ID)
			assert.Equal(t, eventID, status.Events[0].ID)
			assert.True(t, status.TypesKnown)
			assert.Equal(t, 1, status.Events[0].AvailableItems)
			assert.Equal(t, nil, status.Events[0].Err)
			assert.Equal(t, "receiving", status.Events[0].Status)
			assert.Equal(t, "receiving", status.Status)

			status, err = envelopeStorage.StoreEnvelope(env2)
			assert.Nil(t, err)
			assert.Equal(t, env2.ID, status.ID)
			assert.Equal(t, 1, len(status.Events))
			if t.Failed() {
				return
			}
			assert.Equal(t, eventID, status.Events[0].ID)
			assert.Equal(t, 2, status.Events[0].AvailableItems)
			assert.Equal(t, nil, status.Events[0].Err)
			assert.Equal(t, "complete", status.Events[0].Status)
			assert.Equal(t, "complete", status.Status)
		})

		if supportUnordered {
			t.Run("Unordered", func(t *testing.T) {
				envelopeStorage, clean := factory()
				defer clean()

				status, err := envelopeStorage.StoreEnvelope(env2)
				assert.Nil(t, err)
				assert.Equal(t, env1.ID, status.ID)
				assert.Equal(t, "receiving", status.Status)

				status, err = envelopeStorage.StoreEnvelope(env1)
				assert.Nil(t, err)
				assert.Equal(t, env2.ID, status.ID)
				assert.Equal(t, 1, len(status.Events))
				if t.Failed() {
					return
				}
				assert.Equal(t, eventID, status.Events[0].ID)
				assert.True(t, status.TypesKnown)
				assert.Equal(t, 2, status.Events[0].AvailableItems)
				assert.Equal(t, nil, status.Events[0].Err)
				assert.Equal(t, "complete", status.Events[0].Status)
				assert.Equal(t, "complete", status.Status)
			})
		}
	})

	t.Run("Read/SingleEnvelope", func(t *testing.T) {
		envelopeStorage, clean := factory()
		defer clean()
		env := api.Envelope{
			ID: uuid.MustFromString("b74415ee-83c1-11e6-8f04-133b336287df"),
			Events: []api.Event{
				api.Event{
					ID:        uuid.MustFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
					Index:     1,
					Type:      "xbus.test",
					Items:     []api.Item{[]byte("Data")},
					ItemCount: 1,
					Checksum:  1859455373,
				},
			},
		}

		eventID := env.Events[0].ID
		env.EventIDs = []uuid.UUID{eventID}

		status, err := envelopeStorage.StoreEnvelope(env)
		assert.Nil(t, err)
		assert.Equal(t, "complete", status.Status)

		out, pos, err := envelopeStorage.ReadEnvelope(env.ID, storage.EnvelopeStartPos(), 0)
		assert.Nil(t, err)
		assert.True(t, pos.Complete)
		assert.Equal(t, env.ID, out.ID)
		assert.Equal(t, 1, len(out.Events))
		if t.Failed() {
			return
		}
		outEvent := out.Events[0]
		assert.Equal(t, eventID, outEvent.ID)
		assert.Equal(t, 1, outEvent.Index)
		assert.Equal(t, "xbus.test", outEvent.Type)
		assert.Equal(t, []api.Item{[]byte("Data")}, outEvent.Items)
		assert.Equal(t, 1, outEvent.ItemCount)
		assert.Equal(t, api.Checksum(1859455373), outEvent.Checksum)
	})

	t.Run("Stream/PairOfEnvelope", func(t *testing.T) {
		envelopeStorage, clean := factory()
		defer clean()

		env := api.Envelope{
			ID: uuid.MustFromString("b74415ee-83c1-11e6-8f04-133b336287df"),
			Events: []api.Event{
				api.Event{
					ID:    uuid.MustFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
					Index: 1,
					Type:  "xbus.test",
					Items: []api.Item{[]byte("Data")},
				},
			},
		}

		eventID := env.Events[0].ID
		env.EventIDs = []uuid.UUID{eventID}

		status, err := envelopeStorage.StoreEnvelope(env)
		assert.Nil(t, err)
		assert.Equal(t, env.ID, status.ID)
		assert.Equal(t, eventID, status.Events[0].ID)
		assert.True(t, status.TypesKnown)
		assert.Equal(t, 0, status.Events[0].TotalItems)
		assert.Equal(t, 1, status.Events[0].AvailableItems)
		assert.Equal(t, nil, status.Events[0].Err)
		assert.Equal(t, "receiving", status.Events[0].Status)
		assert.Equal(t, "receiving", status.Status)

		out, pos, err := envelopeStorage.ReadEnvelope(env.ID, storage.EnvelopeStartPos(), 0)
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, 1, len(out.Events))
		assert.Equal(t, 1, len(out.Events[0].Items))
		assert.Equal(t, "Data", string(out.Events[0].Items[0]))

		env = api.Envelope{
			ID: env.ID,
			Events: []api.Event{
				api.Event{
					ID:        eventID,
					Index:     2,
					Type:      "xbus.test",
					Items:     []api.Item{[]byte("Data2")},
					ItemCount: 2,
					Checksum:  0x92b3a8e2,
				},
			},
		}

		env.EventIDs = []uuid.UUID{eventID}

		status, err = envelopeStorage.StoreEnvelope(env)
		assert.Nil(t, err)
		assert.Equal(t, env.ID, status.ID)
		assert.Equal(t, 1, len(status.Events))
		if t.Failed() {
			return
		}
		assert.Equal(t, eventID, status.Events[0].ID)
		assert.Equal(t, 2, status.Events[0].AvailableItems)
		assert.Equal(t, nil, status.Events[0].Err)
		assert.Equal(t, "complete", status.Events[0].Status)
		assert.Equal(t, "complete", status.Status)

		out, pos, err = envelopeStorage.ReadEnvelope(env.ID, pos, 0)
		assert.Equal(t, 1, len(out.Events))
		assert.Equal(t, "Data2", string(out.Events[0].Items[0]))
	})

	t.Run("Stream/ItemLessFragments", func(t *testing.T) {
		envID := uuid.MustFromString("b74415ee-83c1-11e6-8f04-133b336287df")
		eventID := uuid.MustFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882")
		eventID2 := uuid.MustFromString("615d146a-02af-11e7-8cdf-d3cb88225aff")

		fragments := []api.Envelope{
			api.Envelope{
				ID:       envID,
				EventIDs: []uuid.UUID{eventID},
				Events: []api.Event{
					api.Event{
						ID:    eventID,
						Index: 1,
						Type:  "xbus.test",
					},
				},
			},
			api.Envelope{
				ID: envID,
				Events: []api.Event{
					api.Event{
						ID:    eventID,
						Index: 2,
						Type:  "xbus.test",
						Items: []api.Item{[]byte("item1")},
					},
				},
			},
			api.Envelope{
				ID: envID,
				Events: []api.Event{
					api.Event{
						ID:    eventID,
						Index: 3,
						Type:  "xbus.test",
					},
				},
			},
			api.Envelope{
				ID: envID,
				Events: []api.Event{
					api.Event{
						ID:        eventID,
						Index:     4,
						ItemCount: 2,
						Type:      "xbus.test",
						Items:     []api.Item{[]byte("item2")},
						Checksum:  0x220c46c7,
					},
				},
			},
		}

		bigEnvFragments := []api.Envelope{
			api.Envelope{
				ID:       envID,
				EventIDs: []uuid.UUID{eventID, eventID2},
				Events: []api.Event{
					api.Event{
						ID:    eventID,
						Index: 1,
						Type:  "xbus.test",
					},
					api.Event{
						ID:    eventID2,
						Index: 1,
						Type:  "xbus.test.2",
					},
				},
			},
			api.Envelope{
				ID: envID,
				Events: []api.Event{
					api.Event{
						ID:    eventID,
						Index: 2,
						Type:  "xbus.test",
					},
				},
			},
			api.Envelope{
				ID: envID,
				Events: []api.Event{
					api.Event{
						ID:    eventID,
						Index: 3,
						Type:  "xbus.test",
					},
				},
			},
			api.Envelope{
				ID: envID,
				Events: []api.Event{
					api.Event{
						ID:    eventID,
						Index: 4,
						Type:  "xbus.test",
					},
				},
			},
		}

		item := make([]byte, 100)

		for i := 0; i != 100; i++ {
			bigEnvFragments = append(bigEnvFragments, api.Envelope{
				ID: envID,
				Events: []api.Event{
					api.Event{
						ID:    eventID,
						Index: i + 5,
						Type:  "xbus.test",
						Items: []api.Item{item},
					},
				},
			})
		}
		for i := 0; i != 100; i++ {
			bigEnvFragments = append(bigEnvFragments, api.Envelope{
				ID: envID,
				Events: []api.Event{
					api.Event{
						ID:    eventID2,
						Index: i + 2,
						Type:  "xbus.test.2",
						Items: []api.Item{item},
					},
				},
			})
		}

		bigEnvFragments = append(bigEnvFragments,
			api.Envelope{
				ID: envID,
				Events: []api.Event{
					api.Event{
						ID:    eventID,
						Index: 105,
						Type:  "xbus.test",
					},
				},
			},
			api.Envelope{
				ID: envID,
				Events: []api.Event{
					api.Event{
						ID:        eventID,
						Index:     106,
						Type:      "xbus.test",
						Items:     []api.Item{item},
						ItemCount: 101,
						Checksum:  0xbdac9720,
					},
					api.Event{
						ID:        eventID2,
						Index:     102,
						Type:      "xbus.test.2",
						Items:     []api.Item{item},
						ItemCount: 101,
						Checksum:  0x560279ee,
					},
				},
			})

		t.Run("Ordered", func(t *testing.T) {
			envelopeStorage, clean := factory()
			defer clean()

			status, err := envelopeStorage.StoreEnvelope(fragments[0])
			assert.Nil(t, err)
			assert.Equal(t, "receiving", status.Status)

			status, err = envelopeStorage.StoreEnvelope(fragments[1])
			assert.Nil(t, err)
			assert.Equal(t, "receiving", status.Status)

			status, err = envelopeStorage.StoreEnvelope(fragments[2])
			assert.Nil(t, err)
			assert.Equal(t, "receiving", status.Status)

			status, err = envelopeStorage.StoreEnvelope(fragments[3])
			assert.Nil(t, err)
			assert.Equal(t, "complete", status.Status)
			assert.Equal(t, "complete", status.Error())

			out, pos, err := envelopeStorage.ReadEnvelope(envID, storage.EnvelopeStartPos(), 0)
			assert.Nil(t, err)
			assert.True(t, pos.Complete)
			assert.Equal(t, 1, len(out.Events))
			if len(out.Events) == 1 {
				assert.Equal(t, []api.Item{api.Item("item1"), api.Item("item2")}, out.Events[0].Items)
			}
		})

		t.Run("Ordered/ReadBig", func(t *testing.T) {
			envelopeStorage, clean := factory()
			defer clean()

			for _, fragment := range bigEnvFragments[:4] {
				status, err := envelopeStorage.StoreEnvelope(fragment)
				assert.Nil(t, err)
				assert.Equal(t, "receiving", status.Status)
			}

			go func() {
				time.Sleep(100 * time.Millisecond)
				for _, fragment := range bigEnvFragments[4 : len(bigEnvFragments)-1] {
					status, err := envelopeStorage.StoreEnvelope(fragment)
					assert.Nil(t, err)
					assert.Equal(t, "receiving", status.Status)
				}
				status, err := envelopeStorage.StoreEnvelope(bigEnvFragments[len(bigEnvFragments)-1])
				assert.Nil(t, err)
				assert.Equal(t, "complete", status.Status)
				assert.Equal(t, "complete", status.Error())

			}()

			pos := storage.EnvelopeStartPos()
			itemCount := 0
			for !pos.Complete {
				out, newPos, err := envelopeStorage.ReadEnvelope(envID, pos, 500)
				if err != nil {
					t.Fatal(err)
				}
				if newPos.Equals(pos) {
					time.Sleep(100 * time.Millisecond)
				}
				for _, ev := range out.Events {
					itemCount += len(ev.Items)
					if newPos.Events[eventID].ItemCount > 202 {
						t.Fatal("ReadEnvelope returning too many items")
					}
				}
				pos = newPos
			}
			assert.Equal(t, 202, itemCount)
		})

		if supportUnordered {
			t.Run("Unordered", func(t *testing.T) {
				envelopeStorage, clean := factory()
				defer clean()

				status, err := envelopeStorage.StoreEnvelope(fragments[2])
				assert.Nil(t, err)
				assert.Equal(t, "receiving", status.Status)

				status, err = envelopeStorage.StoreEnvelope(fragments[0])
				assert.Nil(t, err)
				assert.Equal(t, "receiving", status.Status)

				status, err = envelopeStorage.StoreEnvelope(fragments[3])
				assert.Nil(t, err)
				assert.Equal(t, "receiving", status.Status)

				status, err = envelopeStorage.StoreEnvelope(fragments[1])
				assert.Nil(t, err)
				assert.Equal(t, "complete", status.Status)

				out, pos, err := envelopeStorage.ReadEnvelope(envID, storage.EnvelopeStartPos(), 0)
				assert.Nil(t, err)
				assert.True(t, pos.Complete)
				assert.Equal(t, 1, len(out.Events))
				if len(out.Events) == 1 {
					assert.Equal(t, []api.Item{api.Item("item1"), api.Item("item2")}, out.Events[0].Items)
				}
			})
		}

	})

}
