// Copyright 2017 Orus.io. All rights reserved.

package xbus

import (
	"errors"
	"fmt"
	"runtime/debug"
	"strings"
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/nats-io/go-nats"

	"bitbucket.org/orus-io/go-xbus/api"
)

// EnvelopeHandler is a type of callbacks for when new envelope arrive on an input
type EnvelopeHandler func(*EnvelopeReceiver)

// RawEnvelopeHandler is a type of callback for envelope handlers
type RawEnvelopeHandler func(input, reply string, envelope api.Envelope)

// newActor initialize an actor
func newActor(client *Client, options ActorOptions) *Actor {
	actor := Actor{
		ID:      options.ID,
		Name:    options.Name,
		Kind:    options.Kind,
		Options: options,
		Client:  client,
	}
	actor.Log = &ActorLogger{
		NewPrefixedLogger("[actor/"+options.Name+"] ", client.Log),
		&actor}
	return &actor
}

// Actor provides the common tools for all actors
type Actor struct {
	ID      uuid.UUID
	Name    string
	Kind    api.ActorKind
	Options ActorOptions
	Log     *ActorLogger
	Client  *Client
}

// GetStringSetting returns a string typed setting entry
func (a *Actor) GetStringSetting(name string) string {
	value, ok := a.Options.Service.Settings[name]
	if !ok {
		return ""
	}
	sValue, ok := value.(string)
	if !ok {
		panic(fmt.Sprintf("Invalid configuration value: %s", value))
	}
	return sValue
}

// MustGetStringSetting returns a string setting or panic if undefined
func (a *Actor) MustGetStringSetting(name string) string {
	value := a.GetStringSetting(name)
	if value == "" {
		a.Client.Log.Fatalf("Missing setting: %s", name)
	}
	return value
}

// GetIntSettingD returns a int setting, or the defaultValue if not specified
func (a *Actor) GetIntSettingD(name string, defaultValue int) int {
	value, ok := a.Options.Service.Settings[name]
	if !ok {
		return defaultValue
	}
	iValue, ok := value.(int)
	if !ok {
		panic(fmt.Sprintf("Invalid configuration value '%s'. Expected an integer", value))
	}
	return iValue
}

// GetClient returns the underlying Client instance
func (a *Actor) GetClient() *Client {
	return a.Client
}

// ProcessingEnd calls the ActorProcessingEnd xbus api
func (a *Actor) ProcessingEnd(
	context api.EnvelopeContext,
	envelopeID uuid.UUID,
	result api.ActorProcessingResultType,
	messages []api.LogMessage,
) error {
	var rep api.ActorProcessingEndRep
	if err := a.Client.GetConn().Request(
		api.ActorProcessingEnd(a.ID),
		api.ActorProcessingEndReq{
			Context:    context,
			EnvelopeID: envelopeID,
			Result:     result,
			Messages:   messages,
		},
		&rep,
		10*time.Second,
	); err != nil {
		return err
	}
	if rep.Status != api.StatusOK {
		return wrapAPIError(rep.Response)
	}
	return nil
}

// ProcessingSuccess signal the success of a processing
// for actors with no outputs of it not all outputs were
// hit or closed
func (a *Actor) ProcessingSuccess(context api.EnvelopeContext, envelopeID uuid.UUID) error {
	return a.ProcessingEnd(context, envelopeID, api.ActorProcessingSuccess, nil)
}

// ProcessingError signals a processing error.
func (a *Actor) ProcessingError(context api.EnvelopeContext, envelopeID uuid.UUID, format string, fmtArgs ...interface{}) error {
	return a.ProcessingEnd(context, envelopeID, api.ActorProcessingError, []api.LogMessage{
		api.LogMessage{
			Time:  time.Now(),
			Level: api.LogError,
			Text:  fmt.Sprintf(format, fmtArgs...),
		}})
}

// SendEnvelope send a potentially incomplete envelope to the broker
func (a *Actor) SendEnvelope(output string, env api.Envelope) (status api.ReceptionStatus, err error) {
	if a.Kind == api.ConsumerKind {
		return api.ReceptionError, fmt.Errorf("A 'Consumer' actor cannot emit to outputs")
	}
	return a.Client.SendEnvelope(a.ID, output, env)
}

// SendItems send a group of items to the broker
// The envelope and event will be constructed automatically
func (a *Actor) SendItems(output, eventType string, items []api.Item) (status api.ReceptionStatus, err error) {
	return a.SendEnvelope(output, ForgeEnvelopeFromItems(eventType, items))
}

// SendItemsContext send a group of items to the broker, in a given context
// The envelope and event will be constructed automatically
func (a *Actor) SendItemsContext(context api.EnvelopeContext, output, eventType string, items []api.Item) (status api.ReceptionStatus, err error) {
	envelope := ForgeEnvelopeFromItems(eventType, items)
	envelope.Context = context
	return a.SendEnvelope(output, envelope)
}

// ForgeEnvelopeFromEvent ...
func ForgeEnvelopeFromEvent(event api.Event) api.Envelope {
	envID, err := uuid.V4()
	if err != nil {
		panic(err)
	}
	return api.Envelope{
		ID:       envID,
		EventIDs: []uuid.UUID{event.ID},
		Events:   []api.Event{event},
	}
}

// ForgeEnvelopeFromItems ...
func ForgeEnvelopeFromItems(eventType string, items []api.Item) api.Envelope {
	evID, err := uuid.V4()
	if err != nil {
		panic(err)
	}
	checksum := api.Checksum(0).
		Update(evID[:16]).
		Update([]byte(eventType))

	for _, item := range items {
		checksum = checksum.Update(item)
	}

	event := api.Event{
		ID:        evID,
		Type:      eventType,
		Index:     1,
		Items:     items,
		ItemCount: len(items),
		Checksum:  checksum,
	}
	return ForgeEnvelopeFromEvent(event)
}

// RawSubscription is what RawSubscribe returns
type RawSubscription interface {
	Unsubscribe() error
}

// RawSubscribe subscribes to the raw envelope stream
// The handler must:
// - validates the envelope integrity itself
// - send a EnvelopeAck to the 'reply' subject.
// If wildcard is false, the handler will have the responsibility to
// subscribe to the envelope dedicated subject to get all the remaining
// fragments.
// If wildcard is true, all the envelope subjects will be subscribed too,
// meaning the handler will receive not only the first but any fragment
// of the envelopes.
// In the case of multiple subscription on the same Actor (in case of
// redondant processes or multiple local handlers), using the wildcard
// implies a carreful handling of parallelism, as a single envelope fragments
// can be sent to different handlers.
func (a *Actor) RawSubscribe(handler RawEnvelopeHandler, wildcard bool) (RawSubscription, error) {
	if a.Kind == api.EmitterKind {
		return nil, fmt.Errorf("A 'Emitter' actor cannot subscribe to inputs")
	}
	var subject string
	if wildcard {
		subject = api.ActorInputSubject(a.ID, ">")
	} else {
		subject = api.ActorInputSubject(a.ID, "*")
	}
	sub, err := a.Client.conn.QueueSubscribe(
		subject,
		a.ID.String(),
		func(subject, reply string, e api.Envelope) {
			sSubject := strings.Split(subject, ".")
			input := sSubject[4]

			handler(input, reply, e)
		})
	if err != nil {
		return nil, err
	}
	if err := a.Client.conn.Publish(
		api.ActorReadySubject(a.ID), api.ActorReadyMsg{},
	); err != nil {

	}
	return sub, nil
}

// Subscribe setup the subscription
func (a *Actor) Subscribe(handler EnvelopeHandler) (*Subscription, error) {
	s := Subscription{
		actor:   a,
		handler: handler,
	}
	sub, err := a.RawSubscribe(s.OnReceiveEnvelope, false)
	if err != nil {
		return nil, err
	}
	s.sub = sub
	return &s, nil
}

// OpenOutput returns a EnvelopeSender
func (a *Actor) OpenOutput(
	context api.EnvelopeContext, output string, eventTypes ...string,
) *EnvelopeSender {
	var es *EnvelopeSender
	es = NewEnvelopeSender(
		context, eventTypes, api.EnvelopeMaxSize,
		func(e api.Envelope) (api.ReceptionStatus, error) {
			return a.SendEnvelope(output, e)
		})
	return es
}

// Subscription holds the nats subcription and call the handler when
// an envelope starts arriving
type Subscription struct {
	actor   *Actor
	handler EnvelopeHandler
	sub     RawSubscription
}

// Unsubscribe unsuscribes the nats handler
func (s *Subscription) Unsubscribe() error {
	return s.sub.Unsubscribe()
}

func sendAck(client *Client, subject string, ack *api.EnvelopeAck) (success bool) {
	success = true
	if r := recover(); r != nil {
		client.Log.Errorf("Caught error: %s\n%s", r, debug.Stack())
		ack.Response = api.ServerError(fmt.Errorf("Caught error: %s", r))
		ack.ReceptionStatus = api.ReceptionError
		ack.Reason = "Server side error"
		success = false
	}
	if err := client.conn.Publish(subject, ack); err != nil {
		client.Log.Errorf("Error replying an envelope Ack: %s", err)
		success = false
	}
	return
}

func (s *Subscription) OnReceiveEnvelope(input, reply string, e api.Envelope) {
	var (
		er  *EnvelopeReceiver
		err error
	)
	ack := api.EnvelopeAck{ID: e.ID}
	defer func() {
		if sendAck(s.actor.Client, reply, &ack) && err == nil {
			go s.handler(er)
		}
	}()

	s.actor.Log.Noticef("Received new envelope head on input %s", input)

	er, err = NewEnvelopeReceiver(input, e)

	if err != nil {
		ack = api.NewEnvelopeAckError(e.ID, err)
	} else {
		ack.Response.Status = api.StatusOK
		ack.ReceptionStatus = er.Status()
	}

	if er.Status() == api.ReceptionReceiving {
		// Envelope was not complete, we need to listen for next messages
		// until envelope is complete or receiver stops receiving

		// TODO replace e.ID.String() with a function of envelope context + ID
		subject := api.ActorInputSubject(s.actor.ID, input) + "." + e.ID.String()
		var sub *nats.Subscription

		// TODO Handle case were no more envelopes are received -> currently
		// leads to a leak and a never closed receiver.
		// For that, we will need to keep track of all receivers and their
		// subscriptions with a 'lastReceived' timestamp for periodic cleanup
		// This list of subs/receivers will also be useful with an actor is taken
		// down so we can signal the receiver that it's over (or reinitialize it
		// after a reconnection).
		sub, err := s.actor.Client.conn.Subscribe(
			subject,
			func(subject, reply string, e api.Envelope) {
				ack := api.EnvelopeAck{
					Response: api.Response{Status: api.StatusOK},
					ID:       e.ID}
				defer sendAck(s.actor.Client, reply, &ack)

				switch status := er.Status(); status {
				case api.ReceptionReceiving:
					err := er.addEnvelope(e)

					if err != nil {
						ack = api.NewEnvelopeAckError(e.ID, err)
					} else {
						ack.Response.Status = api.StatusOK
						ack.ReceptionStatus = er.Status()
					}
				case api.ReceptionAccepted:
					ack = api.NewEnvelopeAckError(
						e.ID, errors.New("Envelope already accepted"))
				case api.ReceptionError:
					ack.ReceptionStatus = api.ReceptionError
					ack.Reason = "Reception was already in error state"
				}

				if er.Status() != api.ReceptionReceiving {
					// Reception is over, we can unsubscribe
					if err := sub.Unsubscribe(); err != nil {
						s.actor.Client.Log.Errorf("Error unsubscribing: %s", err)
					}
				}
			},
		)
		if err != nil {
			ack = api.NewEnvelopeAckError(e.ID, err)
		}
	}
}
