// Copyright 2016 Orus.io. All rights reserved.

package xbus

import (
	"errors"
	"fmt"
	"time"

	"github.com/m4rw3r/uuid"

	"bitbucket.org/orus-io/go-xbus/api"
)

func wrapAPIError(res api.Response) error {
	return fmt.Errorf("%s: %s", res.Status, errors.New(res.StatusMessage))
}

func (c *Client) actorOp(action string, actors []api.Actor) ([]api.Actor, error) {
	var res api.ControlActorResponse
	err := c.conn.Request(api.ControlActorsPrefix+action, api.ControlActorRequest{Actors: actors}, &res, 5*time.Second)
	if err != nil {
		return nil, err
	}
	if res.Status != api.StatusOK {
		return nil, wrapAPIError(res.Response)
	}
	return res.Actors, nil
}

// ActorCreate create actors
func (c *Client) ActorCreate(actors []api.Actor) ([]api.Actor, error) {
	return c.actorOp(api.ActionCreate, actors)
}

// ActorList lists actors.
func (c *Client) ActorList(actors []api.Actor) ([]api.Actor, error) {
	return c.actorOp(api.ActionQuery, actors)
}

// ActorAccept accepts a pending actor
func (c *Client) ActorAccept(id uuid.UUID) (api.Actor, error) {
	var rep api.ControlActorResponse
	if err := c.conn.Request(
		api.ControlActorsUpdate,
		api.ControlActorRequest{
			Actors: []api.Actor{api.Actor{
				ID:     id,
				Status: api.ActorEnabled,
			}},
		},
		&rep, 5*time.Second,
	); err != nil {
		return api.Actor{}, err
	}
	if rep.Status != api.StatusOK {
		return api.Actor{}, wrapAPIError(rep.Response)
	}
	return rep.Actors[0], nil
}

// ActorReject rejects a pending actor
func (c *Client) ActorReject(id uuid.UUID) (api.Actor, error) {
	var rep api.ControlActorResponse
	if err := c.conn.Request(
		api.ControlActorsUpdate,
		api.ControlActorRequest{
			Actors: []api.Actor{api.Actor{
				ID:     id,
				Status: api.ActorRejected,
			}},
		},
		&rep, 5*time.Second,
	); err != nil {
		return api.Actor{}, err
	}
	if rep.Status != api.StatusOK {
		return api.Actor{}, wrapAPIError(rep.Response)
	}
	return rep.Actors[0], nil
}

// GraphQuery searches for graphs
func (c *Client) GraphQuery(name string, active bool) ([]string, error) {
	var rep api.ControlGraphQueryRep
	if err := c.conn.Request(api.ControlGraphQuery, api.ControlGraphQueryReq{
		Name:   name,
		Active: active,
	}, &rep, 5*time.Second); err != nil {
		return nil, err
	}
	if rep.Status != api.StatusOK {
		return nil, wrapAPIError(rep.Response)
	}
	var graphs []string
	for _, graph := range rep.GraphList {
		graphs = append(graphs, graph)
	}
	return graphs, nil
}

// GraphPush pushes a graph to the server
func (c *Client) GraphPush(graph string) (string, []string, error) {
	var rep api.ControlGraphPushRep
	if err := c.conn.Request(api.ControlGraphPush, api.ControlGraphPushReq{Graph: graph}, &rep, 5*time.Second); err != nil {
		return "", nil, err
	}
	if rep.Status != api.StatusOK {
		return "", rep.ValidationMessages, wrapAPIError(rep.Response)
	}
	return rep.Graph, rep.ValidationMessages, nil
}

// ProcessQuery queries processes
func (c *Client) ProcessQuery(includeTerminated bool) ([]api.Process, error) {
	var rep api.ControlProcessQueryRep
	if err := c.conn.Request(
		api.ControlProcessQuery,
		api.ControlProcessQueryReq{IncludeTerminated: includeTerminated},
		&rep, 5*time.Second,
	); err != nil {
		return nil, err
	}
	if rep.Status != api.StatusOK {
		return nil, wrapAPIError(rep.Response)
	}
	return rep.ProcessList, nil
}

// ProcessControl pause/resume/cancel a process
func (c *Client) ProcessControl(processID uuid.UUID, command string) error {
	var rep api.ControlProcessControlRep
	if err := c.conn.Request(
		api.ControlProcessControl,
		api.ControlProcessControlReq{
			ProcessID: processID,
			Command:   command,
		},
		&rep, 5*time.Second,
	); err != nil {
		return err
	}
	if rep.Status != api.StatusOK {
		return wrapAPIError(rep.Response)
	}
	return nil
}

// AccountList lists the bus accounts
func (c *Client) AccountList() ([]api.Account, error) {
	var rep api.ControlAccountListRep
	if err := c.conn.Request(
		api.ControlAccountList,
		api.ControlAccountListReq{},
		&rep, 5*time.Second,
	); err != nil {
		return nil, err
	}
	if rep.Status != api.StatusOK {
		return nil, wrapAPIError(rep.Response)
	}
	return rep.AccountList, nil
}

// AccountAccept accepts a pending account
func (c *Client) AccountAccept(id uuid.UUID) (api.Account, error) {
	var rep api.ControlAccountUpdateRep
	if err := c.conn.Request(
		api.ControlAccountUpdate,
		api.ControlAccountUpdateReq{
			Account: api.Account{
				ID:     id,
				Status: api.AccountStatusEnabled,
			},
		},
		&rep, 5*time.Second,
	); err != nil {
		return api.Account{}, err
	}
	if rep.Status != api.StatusOK {
		return api.Account{}, wrapAPIError(rep.Response)
	}
	return rep.Account, nil
}

// AccountReject rejects a pending account
func (c *Client) AccountReject(id uuid.UUID) (api.Account, error) {
	var rep api.ControlAccountUpdateRep
	if err := c.conn.Request(
		api.ControlAccountUpdate,
		api.ControlAccountUpdateReq{
			Account: api.Account{
				ID:     id,
				Status: api.AccountStatusDisabled,
			},
		},
		&rep, 5*time.Second,
	); err != nil {
		return api.Account{}, err
	}
	if rep.Status != api.StatusOK {
		return api.Account{}, wrapAPIError(rep.Response)
	}
	return rep.Account, nil
}
