// Copyright 2016 Orus.io. All rights reserved.

package xbus

import (
	"crypto"
	"crypto/ecdsa"
	"crypto/tls"
	"crypto/x509"
	"fmt"
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/nats-io/go-nats"
	"github.com/nats-io/nuid"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/tools"
)

// ServiceOptions holds options for an actor service
type ServiceOptions struct {
	Type     string
	Settings map[string]interface{}
}

// ActorOptions holds options for an actor
type ActorOptions struct {
	api.Actor `yaml:",inline"`
	Service   ServiceOptions
}

// COptions configures a Client
type COptions struct {
	ID   uuid.UUID
	Name string
	Type api.AccountType

	PubKey  crypto.PublicKey
	PrivKey *ecdsa.PrivateKey

	Certificate []byte
	CSR         []byte
	RootCAs     [][]byte

	ActorList []ActorOptions

	ConnOptions nats.Options

	// ChangeCallback is a function called whenever a change has occured on this set of options.
	// This happens, in particular, during account registration.
	// Optional.
	ChangeCallback func(COptions)
}

// NewClient creates a new client
func NewClient(options COptions, logger Logger) *Client {
	if options.ChangeCallback == nil {
		options.ChangeCallback = func(COptions) {}
	}
	return &Client{
		options: options,
		Log:     logger,
		actors:  make(map[uuid.UUID]*Actor),
		quit:    make(chan bool),
		stopped: make(chan bool),
	}
}

// Client is a xbus client api
type Client struct {
	options       COptions
	conn          *tools.Conn
	authenticated bool
	Log           Logger
	actors        map[uuid.UUID]*Actor
	running       []ActorService
	quit          chan bool
	stopped       chan bool
}

// GetConn returns the underlying nats connection
func (c *Client) GetConn() *tools.Conn {
	return c.conn
}

// ID returns the account ID (may be zero)
func (c *Client) ID() uuid.UUID {
	return c.options.ID
}

// Connect establises a connection to the bus
func (c *Client) Connect() error {
	return c.connect(nil)
}

func (c *Client) connect(disconnectedCB nats.ConnHandler) error {
	var ns string
	opts := c.options.ConnOptions
	opts.Secure = true
	opts.DisconnectedCB = disconnectedCB

	if c.options.Certificate == nil {
		c.Log.Noticef("Anonymous connection")
		opts.User = nuid.Next()
		ns = "anon-" + opts.User

	} else {
		c.Log.Debugf("Using TLS cert to connect")

		// Fill TLSConfig elements based on upper settings when necessary.
		if opts.TLSConfig == nil {
			opts.TLSConfig = &tls.Config{}
		}
		if opts.TLSConfig.Certificates == nil {
			opts.TLSConfig.Certificates = []tls.Certificate{tls.Certificate{
				Certificate: [][]byte{c.options.Certificate},
				PrivateKey:  c.options.PrivKey,
			}}
		}
		if opts.TLSConfig.RootCAs == nil {
			var rootCAs *x509.CertPool
			if c.options.RootCAs != nil {
				rootCAs = x509.NewCertPool()
				for _, ca := range c.options.RootCAs {
					cert, err := x509.ParseCertificate(ca)
					if err != nil {
						return fmt.Errorf("Cannot parse root ca: %s", err)
					}
					rootCAs.AddCert(cert)
				}
			}
			opts.TLSConfig.RootCAs = rootCAs
		}

		ns = c.options.ID.String()
	}

	conn, err := opts.Connect()
	if err != nil {
		return err
	}
	eConn, err := nats.NewEncodedConn(conn, "json")
	if err != nil {
		return err
	}

	c.conn = &tools.Conn{EncodedConn: eConn, Namespace: ns}

	c.authenticated = c.options.Certificate != nil

	return nil
}

// Close closes the connection to the bus
func (c *Client) Close() {
	c.conn.Close()
	c.conn = nil
}

// AddActor attach an actor
func (c *Client) AddActor(actor ActorOptions) {
	c.options.ActorList = append(c.options.ActorList, actor)
}

// GetActorOptions returns an actor options from its id
func (c *Client) GetActorOptions(id uuid.UUID) (ActorOptions, bool) {
	for _, actor := range c.options.ActorList {
		if actor.ID == id {
			return actor, true
		}
	}
	return ActorOptions{}, false
}

// GetActorOptionsByName returns an actor options from its name
func (c *Client) GetActorOptionsByName(name string) (ActorOptions, bool) {
	for _, actor := range c.options.ActorList {
		if actor.Name == name {
			return actor, true
		}
	}
	return ActorOptions{}, false
}

// getActorFromOptions returns a *Actor from the options
func (c *Client) getActorFromOptions(options ActorOptions) *Actor {
	actor, ok := c.actors[options.ID]
	if !ok {
		actor = newActor(c, options)
		c.actors[options.ID] = actor
	}
	return actor
}

// GetActor returns a *Actor from its ID, or nil if it undefined
func (c *Client) GetActor(id uuid.UUID) *Actor {
	actor, ok := c.actors[id]
	if !ok {
		options, ok := c.GetActorOptions(id)
		if !ok {
			return nil
		}
		actor = c.getActorFromOptions(options)
	}
	return actor
}

// GetActorByName returns an actor from its name
func (c *Client) GetActorByName(name string) *Actor {
	actorOpts, ok := c.GetActorOptionsByName(name)
	if ok {
		actor, ok := c.actors[actorOpts.ID]
		if !ok {
			actor = c.getActorFromOptions(actorOpts)
		}
		return actor
	}
	return nil
}

// StartServices initialize and starts all the actors having a ActorService
func (c *Client) StartServices() error {
	var actors []ActorService
	for _, actorOpts := range c.options.ActorList {
		if actorOpts.Service.Type != "" {
			actors = append(actors, MakeActorService(
				actorOpts.Service.Type, c.getActorFromOptions(actorOpts)))
		}
	}
	for i, actor := range actors {
		if err := actor.Startup(); err != nil {
			c.Log.Errorf("Error starting actor %s: %s", c.options.ActorList[i].Name, err)
			if err := c.StopServices(); err != nil {
				c.Log.Errorf("Error shutting down actors: %s", err)
			}
			return err
		}
		c.running = append(c.running, actor)
	}
	return nil
}

// StopServices stops all the running actor
func (c *Client) StopServices() error {
	var lastError error
	actors := c.running[:]
	for _, actor := range actors {
		c.running = c.running[1:]
		if lastError = actor.Shutdown(); lastError != nil {
			c.Log.Errorf("Error shutting down actor: %s", lastError)
		}
	}
	return lastError
}

// Startup attempts to connect and start the services.
// If the connection attempt fail, it waits a little and
// retries forever.
// It blocks until a first successful connect + services start is performed*
func (c *Client) Startup() error {
	var (
		disconnected = make(chan bool)
	)
	if err := c.connect(func(*nats.Conn) {
		disconnected <- true
	}); err != nil {
		return err
	}
	if err := c.StartServices(); err != nil {
		return err
	}

	go func() {
		defer func() {
			c.stopped <- true
		}()
		var retry = make(chan bool)
		for {
			select {
			case <-disconnected:
				c.Log.Errorf("Got disconnected. Stopping services then reconnect")
				if err := c.StopServices(); err != nil {
					c.Log.Errorf("Error stopping services: %s", err)
				}
				c.Close()
			case <-c.quit:
				if err := c.StopServices(); err != nil {
					c.Log.Errorf("Error stopping services: %s", err)
				}
				if c.conn != nil {
					c.Close()
				}
				return
			case <-retry:
			}
			c.Log.Debugf("Reconnecting")
			if err := c.connect(func(*nats.Conn) {
				disconnected <- true
			}); err != nil {
				c.Log.Errorf("Error reconnecting: %s", err)
				time.AfterFunc(time.Second, func() { retry <- true })
				continue
			}
			c.Log.Debugf("Start services")
			err := c.StartServices()
			if err != nil {
				c.Log.Errorf("Error starting services: %s", err)
				c.Close()
				time.AfterFunc(time.Second, func() { retry <- true })
			}
		}
	}()
	return nil
}

// Shutdown stops the services if running and disconnect from
// the bus
func (c *Client) Shutdown() {
	c.quit <- true
	<-c.stopped
}
